from collections import namedtuple

import numpy as np
from tqdm.auto import tqdm

from openpnm._skgraph.simulations import (
    bond_percolation,
    find_connected_clusters,
    site_percolation,
)
from openpnm.algorithms import Algorithm
from openpnm.utils import Docorator, TypedSet

docstr = Docorator()


__all__ = ['Drainage']


@docstr.get_sections(base='DrainageSettings',
                     sections=['Parameters'])
@docstr.dedent
class DrainageSettings:
    r"""
    Parameters
    ----------
    %(AlgorithmSettings.parameters)s

    throat_entry_pressure : str
        The dictionary key for the pore entry pressure array
    pore_volume : str
        The dictionary key for the pore volume array
    throat_volume : str
        The dictionary key for the throat volume array

    """
    phase = ''
    throat_entry_pressure = 'throat.entry_pressure'
    pore_volume = 'pore.volume'
    throat_volume = 'throat.volume'
    variable_props = TypedSet()


class Drainage(Algorithm):
    """A class to simulate drainage."""

    def __init__(self, phase, name='drainage_?', **kwargs):
        super().__init__(name=name, **kwargs)
        self.settings._update(DrainageSettings())
        self.settings['phase'] = phase.name
        self['pore.bc.inlet'] = False
        self['pore.bc.outlet'] = False
        self.reset()

    def reset(self):
        r"""
        Resets the algorithm's main results so that it can be re-run
        """
        self['pore.invaded'] = False
        self['throat.invaded'] = False
        # self['pore.residual'] = False
        # self['throat.residual'] = False
        self['pore.trapped'] = False
        self['throat.trapped'] = False
        self['pore.invasion_pressure'] = np.inf
        self['throat.invasion_pressure'] = np.inf
        self['pore.invasion_sequence'] = -1
        self['throat.invasion_sequence'] = -1

    def _set_residual(self, pores=None, throats=None, mode='add'):  # pragma: no cover
        raise NotImplementedError("The ability to add residual nwp is not ready yet")
        if pores is not None:
            self['pore.invaded'][pores] = True
            self['pore.residual'][pores] = True
            self['pore.invasion_pressure'][self['pore.invaded']] = -np.inf
            self['pore.invasion_sequence'][pores] = 0
        if throats is not None:
            self['throat.invaded'][throats] = True
            self['throat.residual'][throats] = True
            self['throat.invasion_pressure'][self['throat.invaded']] = -np.inf
            self['throat.invasion_sequence'][throats] = 0

    def set_inlet_BC(self, pores, mode='add'):
        r"""
        Specify the pores from which invading fluid will enter the domain

        Parameters
        ----------
        pores : ndarray
            List of pore indices
        mode : str
            How the boundary conditions should be applied. Options are:

            ============ =====================================================
            mode         meaning
            ============ =====================================================
            'add'        (default) Adds the supplied boundary conditions to
                         the given locations. Raises an exception if values
                         of any type already exist in the given locations.
            'overwrite'  Adds supplied boundary conditions to the given
                         locations, including overwriting conditions of the
                         given type or any other type that may be present in
                         the given locations.
            'remove'     Removes boundary conditions of the specified type
                         from the specified locations. If ``bctype`` is not
                         specified then *all* types are removed. If no
                         locations are given then values are remvoed from
                         *all* locations.
            ============ =====================================================
        """
        self.set_BC(pores=pores, bcvalues=True, bctype='inlet', mode=mode)

    def set_outlet_BC(self, pores, mode='add'):
        r"""
        Specify the pores from which defending fluid exits the network.

        This is optional and only required if trapping is to be considered.

        Parameters
        ----------
        pores : ndarray
            The list of outlet pores
        mode : str
            How the boundary conditions should be applied. Options are:

            ============ =====================================================
            mode         meaning
            ============ =====================================================
            'add'        (default) Adds the supplied boundary conditions to
                         the given locations. Raises an exception if values
                         of any type already exist in the given locations.
            'overwrite'  Adds supplied boundary conditions to the given
                         locations, including overwriting conditions of the
                         given type or any other type that may be present in
                         the given locations.
            'remove'     Removes boundary conditions of the specified type
                         from the specified locations. If ``bctype`` is not
                         specified then *all* types are removed. If no
                         locations are given then values are remvoed from
                         *all* locations.
            ============ =====================================================
        """
        self.set_BC(pores=pores, bcvalues=True, bctype='outlet', mode=mode)

    def run(self, pressures=25):
        r"""
        Runs the simulation for the pressure points

        Parameters
        ----------
        pressures : int or ndarray
            The number of pressue steps to apply, or an array of specific
            points

        """
        if isinstance(pressures, int):
            phase = self.project[self.settings.phase]
            hi = 1.25*phase[self.settings.throat_entry_pressure].max()
            low = 0.80*phase[self.settings.throat_entry_pressure].min()
            pressures = np.logspace(np.log10(low), np.log10(hi), pressures)
        pressures = np.array(pressures, ndmin=1)
        msg = 'Performing drainage simulation'
        for i, p in enumerate(tqdm(pressures, msg)):
            self._run_special(p)
            pmask = self['pore.invaded'] * (self['pore.invasion_pressure'] == np.inf)
            self['pore.invasion_pressure'][pmask] = p
            self['pore.invasion_sequence'][pmask] = i
            tmask = self['throat.invaded'] * (self['throat.invasion_pressure'] == np.inf)
            self['throat.invasion_pressure'][tmask] = p
            self['throat.invasion_sequence'][tmask] = i
        # If any outlets were specified, evaluate trapping
        if np.any(self['pore.bc.outlet']):
            self.apply_trapping()

    def _run_special(self, pressure):
        phase = self.project[self.settings.phase]
        Tinv = phase[self.settings.throat_entry_pressure] <= pressure
        # Tinv += self['throat.invaded']
        # Remove trapped throats from this list, if any
        # Tinv[self['throat.trapped']] = False
        # Perform bond_percolation to label invaded clusters
        s_labels, b_labels = bond_percolation(self.network.conns, Tinv)
        # Remove label from any clusters not connected to the inlets
        s_labels, b_labels = find_connected_clusters(
            b_labels, s_labels, self['pore.bc.inlet'], asmask=False)
        # Add result to existing invaded locations
        self['pore.invaded'][s_labels >= 0] = True
        self['throat.invaded'][b_labels >= 0] = True

    def apply_trapping(self):
        r"""
        Adjusts the invasion history of pores and throats that are trapped.

        Returns
        -------
        This function returns nothing, but the following adjustments are made
        to the data on the object for trapped pores and throats:

        * ``'pore/throat.trapped'`` is set to ``True``
        * ``'pore/throat.invaded'`` is set to ``False``
        * ``'pore/throat.invasion_pressure'`` is set to ``np.inf``
        * ``'pore/throat.invasion_sequence'`` is set to ``0``

        Notes
        -----
        This search proceeds by the following 3 steps:

        1. A site percolation is applied to *uninvaded* pores and they are set
        to trapped if they belong to a cluster that is not connected to the
        outlets.

        2. All throats which were invaded at a pressure *higher* than either
        of its two neighboring pores are set to trapped, regardless of
        whether the pores themselves are trapped.

        3. All throats which are connected to trapped pores are set to trapped
        as these cannot be invaded since the fluid they contain cannot escape.

        """
        pseq = self['pore.invasion_pressure']
        tseq = self['throat.invasion_pressure']
        # Firstly, find any throats who were invaded at a pressure higher than
        # either of its two neighboring pores
        temp = (pseq[self.network.conns].T > tseq).T
        self['throat.trapped'][np.all(temp, axis=1)] = True
        # Now scan through and use site percolation to find other trapped
        # clusters of pores
        for p in np.unique(pseq):
            s, b = site_percolation(conns=self.network.conns,
                                    occupied_sites=pseq > p)
            # Identify uninvaded throats between previously invaded pores within same cluster   
            both_pores_invaded = (pseq[self.network.conns[:, 0]] <= p) & (pseq[self.network.conns[:, 1]] <= p)
            same_cluster = s[self.network.conns[:, 0]] == s[self.network.conns[:, 1]]
            uninvaded_throat = tseq > p
            trap_condition = both_pores_invaded & same_cluster & uninvaded_throat
            self['throat.trapped'][trap_condition] = True
            # Identify cluster numbers connected to the outlets
            clusters = np.unique(s[self['pore.bc.outlet']])
            # Find ALL throats connected to any trapped site, since these
            # throats must also be trapped, and update their cluster numbers
            Ts = self.network.find_neighbor_throats(pores=s >= 0)
            b[Ts] = np.amax(s[self.network.conns], axis=1)[Ts]
            # Finally, mark pores and throats as trapped if their cluster
            # numbers are NOT connected to the outlets
            self['pore.trapped'] += np.isin(s, clusters, invert=True)*(s >= 0)
            self['throat.trapped'] += np.isin(b, clusters, invert=True)*(b >= 0)
        # Use the identified trapped pores and throats to update the other
        # data on the object accordingly
        # self['pore.trapped'][self['pore.residual']] = False
        # self['throat.trapped'][self['throat.residual']] = False
        self['pore.invaded'][self['pore.trapped']] = False
        self['throat.invaded'][self['throat.trapped']] = False
        self['pore.invasion_pressure'][self['pore.trapped']] = np.inf
        self['throat.invasion_pressure'][self['throat.trapped']] = np.inf
        self['pore.invasion_sequence'][self['pore.trapped']] = -1
        self['throat.invasion_sequence'][self['throat.trapped']] = -1

    def pc_curve(self, pressures=None):
        if pressures is None:
            pressures = np.unique(self['pore.invasion_pressure'])
        elif isinstance(pressures, int):
            p = np.unique(self['pore.invasion_pressure'])
            p = p[np.isfinite(p)]
            pressures = np.logspace(np.log10(p.min()/2), np.log10(p.max()*2), pressures)
        else:
            pressures = np.array(pressures)
        pc = []
        s = []
        Vp = self.network[self.settings.pore_volume]
        Vt = self.network[self.settings.throat_volume]
        for p in pressures:
            Snwp_p = self['pore.invasion_pressure'] <= p
            Snwp_t = self['throat.invasion_pressure'] <= p
            pc.append(p)
            s.append(((Snwp_p*Vp).sum() + (Snwp_t*Vt).sum())/(Vp.sum() + Vt.sum()))
        pc_curve = namedtuple('pc_curve', ('pc', 'snwp'))
        data = pc_curve(np.array(pc), np.array(s))
        return data


# %%
# def run_examples():
if __name__ == '__main__':
    import matplotlib.pyplot as plt

    import openpnm as op
    plt.rcParams['figure.facecolor'] = 'darkgrey'
    plt.rcParams['axes.facecolor'] = 'grey'

    np.random.seed(0)
    Nx, Ny, Nz = 10, 10, 1
    pn = op.network.Cubic([Nx, Ny, Nz], spacing=1e-5)
    pn.add_model_collection(op.models.collections.geometry.spheres_and_cylinders)
    pn.regenerate_models()
    nwp = op.phase.Phase(network=pn)
    nwp['throat.surface_tension'] = 0.480
    nwp['throat.contact_angle'] = 140
    nwp.add_model(propname='throat.entry_pressure',
                  model=op.models.physics.capillary_pressure.washburn)
    nwp.add_model(propname='pore.entry_pressure',
                  model=op.models.physics.capillary_pressure.washburn,
                  contact_angle=140,
                  surface_tension=0.480,
                  diameter='pore.diameter')

    # %%
    drn = Drainage(network=pn, phase=nwp)
    drn.set_inlet_BC(pores=pn.pores('left'))
    pressures = np.logspace(np.log10(0.1e6), np.log10(8e6), 40)
    drn.run(pressures)
    drn.set_outlet_BC(pores=pn.pores('right'))
    drn.apply_trapping()

    # %%
    if 1:
        fig, ax = plt.subplots(1, 1, figsize=[20, 20])
        ax.semilogx(*drn.pc_curve(pressures), 'ro-')
        ax.set_ylim([-.05, 1.05])

    if 1:
        pressures = np.unique(drn['pore.invasion_pressure'])
        n = 6
        p = drn['pore.invasion_pressure'] <= pressures[n]
        t = drn['throat.invasion_pressure'] <= pressures[n]
        ax = op.topotools.plot_coordinates(pn, pores=p, s=500,
                                           color_by=drn['pore.invasion_pressure'])
        ax = op.topotools.plot_connections(pn, throats=t, linewidth=5,
                                           color_by=drn['throat.invasion_pressure'],
                                           ax=ax)
        # t = drn['throat.invasion_pressure'] <= p
        # ax = op.topotools.plot_connections(pn, throats=t, c='k', linestyle='--', ax=ax)

    if 0:
        import openpnm as op
        tseq = drn['throat.invasion_pressure']
        pseq = drn['pore.invasion_pressure']
        Pmax = np.amax(tseq[tseq < np.inf])
        pseq[pseq > Pmax] = Pmax + 1
        tseq[tseq > Pmax] = Pmax + 1
        for j, i in enumerate(tqdm(np.unique(tseq)[:-1])):
            ax = op.topotools.plot_connections(pn, tseq <= i, linestyle='--',
                                               c='r', linewidth=3)
            op.topotools.plot_coordinates(pn, pseq <= i, c='b', marker='x',
                                          markersize=100, ax=ax)
            op.topotools.plot_coordinates(pn, drn['pore.trapped'], c='c',
                                          marker='o', markersize=100, ax=ax)
            plt.savefig(f"{str(j).zfill(3)}.png")
            plt.close()

    # %%
    if 0:
        drn = Drainage(network=pn, phase=nwp)
        drn.set_inlets(pores=pn.pores('left'))
        pressures = np.logspace(np.log10(0.1e6), np.log10(8e6), 40)
        drn.run(pressures)
        drn.set_outlets(pores=pn.pores('right'))
        pressures = np.unique(drn['pore.invasion_pressure'])
        pseq = drn['pore.invasion_pressure']
        p = pressures[4]
        s, b = site_percolation(conns=pn.conns, occupied_sites=pseq > p)
        clusters = np.unique(s[drn['pore.bc.outlet']])
        Ts = pn.find_neighbor_throats(pores=s >= 0)
        b[Ts] = np.amax(s[pn.conns], axis=1)[Ts]
        drn['pore.trapped'] += np.isin(s, clusters, invert=True)*(s >= 0)
        drn['throat.trapped'] += np.isin(b, clusters, invert=True)*(b >= 0)
        ax = op.topotools.plot_coordinates(pn, pores=drn['pore.trapped'],
                                           color_by=s)
        ax = op.topotools.plot_coordinates(pn, pores=pseq <= p, c='k', ax=ax)
        ax = op.topotools.plot_connections(pn, throats=drn['throat.trapped'],
                                           color_by=b, ax=ax)
        t = drn['throat.invasion_pressure'] <= p
        ax = op.topotools.plot_connections(pn, throats=t, c='k', linestyle='--', ax=ax)



import logging

import numpy as np

from openpnm.models.collections.phase import (
    binary_gas_mixture,
    standard_gas_mixture,
    standard_liquid_mixture,
)
from openpnm.models.phase.mixtures import mole_summation
from openpnm.phase import Phase
from openpnm.utils import (
    Docorator,
    HealthDict,
    Workspace,
    get_printable_labels,
    get_printable_props,
)

logger = logging.getLogger(__name__)
docstr = Docorator()
ws = Workspace()


__all__ = [
    'Mixture',
    'BinaryGas',
    'GasMixture',
    'LiquidMixture',
    'StandardGasMixture',
    'StandardLiquidMixture',
]


class MixtureSettings:
    ...


@docstr.get_sections(base='Mixture', sections=['Parameters'])
@docstr.dedent
class Mixture(Phase):
    r"""
    Creates ``Phase`` object that represents a multicomponent mixture
    consisting of a given list of individual ``Species`` objects as
    components.

    Parameters
    ----------
    %(Phase.parameters)s
    components : list of ``Species`` objects
        A list of all components that constitute the mixture

    """

    def __init__(self, components=[], name='mixture_?', **kwargs):
        self._components = []
        super().__init__(name=name, **kwargs)
        self.settings._update(MixtureSettings())

        # Add any supplied phases to the list components
        self.components = components

    def __getitem__(self, key):
        try:
            vals = super().__getitem__(key)
        except KeyError:
            # If key ends in component name, fetch it directly
            if key.split('.')[-1] in self.components.keys():
                comp = self.project[key.split('.')[-1]]
                vals = comp[key.rsplit('.', maxsplit=1)[0]]
            elif key[-1] in ['*', '.']:  # Fetch values for ALL components
                key = key[:-1]
                while key.endswith('.'):  # If ended with .* also remove .
                    key = key[:-1]
                vals = self.get_comp_vals(key)
            else:
                raise KeyError(key)
        return vals

    def regenerate_models(self, **kwargs):
        for c in self.components.values():
            c.regenerate_models(**kwargs)
        super().regenerate_models(**kwargs)

    def get_comp_vals(self, propname):
        r"""
        Get a dictionary of the requested values from each component in the
        mixture

        Parameters
        ----------
        propname : str
            The property to fetch, such as ``'pore.viscosity'``.

        Returns
        -------
        vals : dict
            A dictionary with each component name as the key and the requested
            property as the value.

        """
        if not isinstance(propname, str):
            return propname
        if propname.endswith('*'):
            try:
                return self[propname]
            except KeyError:
                pass
        try:
            vals = {}
            for comp in self.components.values():
                vals[comp.name] = comp[propname]
            return vals
        except KeyError:
            msg = f'{propname} not found on at least one component'
            raise Exception(msg)

    def get_mix_vals(self, propname, mode='linear', power=1):
        r"""
        Get the mole fraction weighted value of a given property for the mixture

        Parameters
        ----------
        propname : str
            The property to fetch, such as ``'pore.viscosity'``.
        mode : str
            The type of weighting to use. Options are:

            ============== ====================================================
            mode
            ============== ====================================================
            'linear'       (default) Basic mole fraction weighting of the form:
                           :math:`z = \Sigma (x_i \cdot \z_i)`
            'logarithmic'  Uses the natural logarithm of the property as:
                           :math:`ln(z) = \Sigma (x_i \cdot ln(\z_i))`
            'power         Applies an exponent to the property as:
                           :math:`\z^{power} = \Sigma (x_i \cdot \z_i^{power})`
            ============== ====================================================

        power : scalar
            If ``mode='power'`` this indicates the value of the exponent,
            otherwise this is ignored.

        Returns
        -------
        vals : ndarray
            An ndarray with the requested values obtained using the specified
            weigthing.

        """
        Xs = self['pore.mole_fraction']
        ys = self.get_comp_vals(propname)
        if mode == 'linear':
            z = np.vstack([Xs[k]*ys[k] for k in Xs.keys()]).sum(axis=0)
        elif mode == 'logarithmic':
            z = np.vstack([Xs[k]*np.log(ys[k]) for k in Xs.keys()]).sum(axis=0)
            z = np.exp(z)
        elif mode == 'power':
            z = np.vstack([Xs[k]*(ys[k])**power for k in Xs.keys()]).sum(axis=0)
            z = z**(1/power)
        return z

    @property
    def info(self):  # pragma: no cover
        hr = '―' * 78
        lines = '\n' + 'Component Phases'
        for item in self.components.values():
            lines += '\n' + "═"*78 + '\n' + item.__repr__() + '\n' + hr + '\n'
            lines += get_printable_props(item, suffix=item.name)
            lines += '\n' + hr + '\n'
            lines += get_printable_labels(item, suffix=item.name)
        temp = super().__str__().split(hr)
        temp.insert(-1, lines + '\n')
        lines = hr.join(temp)
        print(lines)

    def _get_comps(self):
        comps = {}
        for k in self.keys():
            if k.startswith('pore.mole_fraction'):
                name = k.split('.')[-1]
                comps[name] = self.project[name]
        return comps

    def _set_comps(self, components):
        if not isinstance(components, list):
            components = [components]
        # Add mole_fraction array to dict, filled with nans
        for item in components:
            if 'pore.mole_fraction.' + item.name not in self.keys():
                self['pore.mole_fraction.' + item.name] = np.nan

    components = property(fget=_get_comps, fset=_set_comps)

    def add_comp(self, component, mole_fraction=np.nan):
        r"""
        Helper method to add a component

        Parameters
        ----------
        component : Species object
            The pure ``Species`` object to add
        mole_fraction : float or array_like
            The mole fraction of the given component in each pore. Can either
            be a scalar which is applied to every pore, or an Np-long
            ``ndarray`` with values for each pore.

        Notes
        -----
        This method just adds `'pore.mole_fraction.<component.name>'`
        to the `Mixture` dictionary, and assigns the given `mole_fraction`.
        """
        self['pore.mole_fraction.' + component.name] = mole_fraction

    def remove_comp(self, component):
        r"""
        Helper method to remove a component from the mixture

        Parameters
        ----------
        component : Species object or str name
            The `Species` to remove from the mixture. Can either be a handle to
            the object, or the object's `name`.

        Notes
        -----
        This method just calls
        `del mixture['pore.mole_fraction.<component.name>']` to remove the
        given component.

        """
        if hasattr(component, 'name'):
            component = component.name
        try:
            del self['pore.mole_fraction.' + component]
        except KeyError:
            pass

    def check_mixture_health(self):
        r"""
        Checks the state of health of the mixture

        Calculates the mole fraction of all species in each pore and returns
        an list of where values are too low or too high,

        Returns
        -------
        health : dict
            A HealthDict object containing lists of locations where the mole
            fractions are not unity. One value indicates locations that are
            too high, and another where they are too low. This `dict` evaluates
            to `True` if all items are healthy.

        """
        h = HealthDict()
        h['mole_fraction_too_low'] = []
        h['mole_fraction_too_high'] = []
        conc = mole_summation(phase=self)
        lo = np.where(conc < 1.0)[0]
        hi = np.where(conc > 1.0)[0]
        if len(lo) > 0:
            h['mole_fraction_too_low'] = lo
        if len(hi) > 0:
            h['mole_fraction_too_high'] = hi
        return h


class LiquidMixture(Mixture):
    r"""
    Wrapper class for creating a liquid mixture.

    This class has a helper method `x` which allows for setting and getting
    of compositions.
    """

    def x(self, compname=None, x=None):
        r"""
        Helper method for getting and setting mole fractions of a component

        Parameters
        ----------
        compname : str, optional
            The name of the component, i.e. ``obj.name``.  If ``x`` is not
            provided this will *return* the mole fraction of the requested
            component.  If ``x`` is provided this will *set* the mole fraction
            of the specified component to ``x``.
        x : scalar or ndarray, optional
            The mole fraction of the given species in the mixture. If not
            provided this method works as a *getter* and will return the
            mole fraction of the requested component.  If ``compname`` is not
            provided then the mole fractions of all components will be returned
            as a dictionary with the components names as keys.

        Notes
        -----
        This method is equivalent to
        ``mixture['pore.mole_fraction.<compname>'] = x``
        """
        if hasattr(compname, 'name'):
            compname = compname.name
        if x is None:
            if compname is None:
                return self['pore.mole_fraction']
            else:
                self['pore.mole_fraction' + '.' + compname]
        else:
            self['pore.mole_fraction.' + compname] = x


class GasMixture(Mixture):
    r"""
    Wrapper class for creating a gas mixture.

    This class has a helper method `y` which allows for setting and getting
    of compositions.
    """

    def y(self, compname=None, y=None):
        r"""
        Helper method for getting and setting mole fractions of a component

        Parameters
        ----------
        compname : str, optional
            The name of the component i.e. ``obj.name``.  If ``y`` is not
            provided this will *return* the mole fraction of the requested
            component.  If ``y`` is provided this will *set* the mole fraction
            of the specified component to ``y``.
        y : scalar or ndarray, optional
            The mole fraction of the given species in the mixture. If not
            provided this method works as a *getter* and will return the
            mole fraction of the requested component. If ``compname`` is also
            not provided then the mole fractions of all components will be
            returned as a dictionary with the components names as keys.

        Notes
        -----
        This method is equivalent to
        ``mixture['pore.mole_fraction.<compname>'] = y``
        """
        if hasattr(compname, 'name'):
            compname = compname.name
        if y is None:
            if compname is None:
                return self['pore.mole_fraction']
            else:
                return self['pore.mole_fraction' + '.' + compname]
        else:
            self['pore.mole_fraction.' + compname] = y


class BinaryGas(GasMixture):
    r"""
    A ``GasMixture`` class that has a predefined set of models for computing the
    properties of the mixture as a function of composition.

    The number of components is capped at 2 (i.e. binary) since the calculation
    of diffusion coefficients using the Lennard-Jones method in the `chemicals`
    package only works for binary mixtures.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_model_collection(binary_gas_mixture)
        # Running models doesn't make sense as compositions are not set yet
        # self.regenerate_models()

    def __setitem__(self, key, value):
        if key.startswith('pore.mole_fraction'):
            if key not in self.keys():
                try:
                    d = self['pore.mole_fraction'].keys()
                    if len(d) >= 2:
                        msg = "Binary mixtures cannot have more than" \
                            " 2 components, remove one first"
                        raise Exception(msg)
                except KeyError:
                    pass
        super().__setitem__(key, value)


class StandardLiquidMixture(LiquidMixture):
    r"""
    A ``LiquidMixture`` class that also has a predefined suite of models for
    computing the properties of the mixture as a function of composition.

    The models are taken from the `chemicals` package and can be viewed with
    ``print(liq_mix.models)``, where `liq_mix` is an instance of this class.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_model_collection(standard_liquid_mixture)
        # Running models doesn't make sense as compositions are not set yet
        # self.regenerate_models()


class StandardGasMixture(GasMixture):
    r"""
    A ``GasMixture`` class that also has a predefined suite of models for
    computing the properties of the mixture as a function of composition.

    The models are taken from the `chemicals` package and can be viewed with
    ``print(gas_mix.models)``, where `gas_mix` is an instance of this class.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_model_collection(standard_gas_mixture)
        # Running models doesn't make sense as compositions are not set yet
        # self.regenerate_models()



from openpnm.models.collections.phase import mercury
from openpnm.phase import Phase, _fetch_chemical_props


__all__ = [
    'Mercury',
]


class Mercury(Phase):
    r"""
    Creates Phase object with and preset values and pore-scale models for
    mercury

    Parameters
    ----------
    %(Phase.parameters)s

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        from thermo import Chemical
        a = Chemical('hg')
        temp = _fetch_chemical_props(a)
        self.params.update(temp)
        self.add_model_collection(mercury)
        self.regenerate_models()



import numpy as np
from openpnm.models.phase import _phasedocs


__all__ = [

]


@_phasedocs
def liquid_mixture_Vc_XXX(
    phase,
    Vcs='pore.critical_volume.*',
):  # pragma: no cover
    r"""
    Calculates the critical volume of a liquid mixture using the correlation
    in [1]_

    Parameters
    ----------
    %(phase)s
    %(Vcs)s

    Returns
    -------

    References
    ----------
    .. [1] tbd

    """
    raise NotImplementedError("This function is not ready yet")
    xs = phase['pore.mole_fraction']
    Vcs = phase.get_comp_vals(Vcs)
    N = len(xs)  # Number of components
    Vm1 = np.sum([xs[i]*Vcs[i] for i in range(N)], axis=0)
    Vm2 = np.sum([xs[i]*(Vcs[i])**(2/3) for i in range(N)], axis=0)
    Vm3 = np.sum([xs[i]*(Vcs[i])**(1/3) for i in range(N)], axis=0)
    Vm = 0.25*(Vm1 + 3*(Vm2*Vm3))
    return Vm


@_phasedocs
def liquid_mixture_Tc_XXX(
    phase,
    Vm='pore.molar_volume',
    Vcs='pore.critical_volume.*',
    Tcs='pore.critical_temperature.*',
):  # pragma: no cover
    r"""
    Calculates the critical temperature of a liquid mixture using the
    correlation in [1]_

    Parameters
    ----------
    %(phase)s
    %(Vm)s
    %(Vcs)s
    %(Tcs)s

    Returns
    -------

    References
    ----------
    .. [1] tbd

    """
    raise NotImplementedError("This function is not ready yet")
    xs = phase['pore.mole_fraction']
    Tcs = phase.get_comp_vals(Tcs)
    Vcs = phase.get_comp_vals(Vcs)
    Vm = phase[Vm]
    N = len(xs)  # Number of components
    num = np.zeros_like(xs[0])
    for i in range(N):
        for j in range(N):
            VT = (Vcs[i]*Tcs[i]*Vcs[j]*Tcs[j])**0.5
            num += xs[i]*xs[j]*VT
    Tcm = num/Vm
    return Tcm


def liquid_mixture_acentric_factor_XXX(
    phase,
    omegas='param.acentric_factor.*',
):  # pragma: no cover
    r"""
    Calculates the accentric factor of a liquid mixture using the correlation
    in [1]_

    Parameters
    ----------
    %(phase)s
    %(omegas)s

    Returns
    -------

    References
    ----------
    .. [1] tbd

    """
    raise NotImplementedError("This function is not ready yet")
    xs = phase['pore.mole_fraction']
    omegas = phase.get_comp_vals(omegas)
    omega = np.sum([omegas[i]*xs[i] for i in range(len(xs))], axis=0)
    return omega



r"""
Miscellaneous
=============

This submodule contains models for calculating general or generic values
that are useful for all other pore-scale models, such as scaling values, or
generating an array of random numbers.

"""

# The following bits are to initialize some boilerplate docstrings for docrep
from openpnm.utils import Docorator as _doc
_docstr = _doc()
_docstr.params['models.misc.seeds'] = \
    r"""seeds : str, optional
            Name of the dictionary key on ``target`` where the array containing
            random seed values (between 0 and 1) is stored. These values are
            used to do a reverse lookup on the cdf of given distribution using
            the ``ppf`` method on the scipy.stats object.  Truncating the
            range (e.g. from 0.1 to 0.9) is often useful to prevent the
            occurance of extreme values from the long tails."""


from ._statistical_distributions import *
from ._simple_equations import *
from ._basic_math import *
from ._neighbor_lookups import *



r"""
Network
=======

Contains network generators and the basic Network class

"""

from ._network import Network
from ._cubic import Cubic
from ._demo import Demo
from ._bcc import BodyCenteredCubic
from ._fcc import FaceCenteredCubic
from ._cubic_template import CubicTemplate
from ._delaunay_voronoi_dual import DelaunayVoronoiDual
from ._voronoi import Voronoi
from ._delaunay import Delaunay



r"""
Throat Perimeter Models
-----------------------

This model contains a selection of functions for calcuating the
throat perimeter assuming a circular/square/rectangular cross-section

"""

from ._funcs import *



import numpy as _np
from openpnm.models.geometry import _geodocs


__all__ = ["pore_coords"]


@_geodocs
def pore_coords(
    network
):
    r"""
    Calculate throat centroid values by averaging adjacent pore coordinates

    Parameters
    ----------
    %(network)s

    Returns
    -------
    values : ndarray
        A numpy ndarray containing throat centroid values

    """
    conns = network['throat.conns']
    coords = network['pore.coords']
    return _np.mean(coords[conns], axis=1)



r"""
Hydraulic Size Factors Models
-----------------------------

This model contains a selection of functions for Computing hydraulic size factors
for conduits assuming pores and throats have various shapes.

"""

from ._funcs import *



import logging

import numpy as np

import openpnm.models.misc as misc
from openpnm.phase import Phase as Phase
from openpnm.utils import Docorator, TypedSet

logger = logging.getLogger(__name__)
docstr = Docorator()


__all__ = [
    'MultiPhase',
    'multiphase_diffusion',
]


@docstr.dedent
class MultiPhaseSettings:
    """
    Parameters
    ----------
    %(PhaseSettings.parameters)s
    phases : list of strings
        The name of the phase objects which comprize this multiphase object
    throat_occupancy : str
        Indicates how throat occupancy is determined.  Options are:

        =========== ==========================================================
        mode        description
        =========== ==========================================================
        'automatic' The occupancy of each throat is calculated based on the
                    occupancy of the two neighboring pores
        'manual'    The occupancy of each throat must be set by hand by
                    assigning values to 'throat.occupancy.{phase.name}'
        =========== ==========================================================

    partition_coef_prefix : str
        The throat property which contains the partition coefficient values
    """
    phases = TypedSet(types=[str])
    throat_occupancy = 'manual'
    partition_coef_prefix = 'throat.partition_coef'


@docstr.dedent
class MultiPhase(Phase):
    """
    Creates a Phase object that represents a multiphase system consisting of
    a given list of Phases.

    Parameters
    ----------
    phases : list[Phase]
        A list containing the phase objects that make up the multiphase system
    %(Phase.parameters)s

    Notes
    -----
    This class assumes that only a SINGLE phase exists in each pore/throat.

    """

    def __init__(self, phases=[], name='mphase_?', **kwargs):
        super().__init__(name=name, **kwargs)
        self.settings._update(MultiPhaseSettings())

        # Pressure/temperature must come from constituent phases
        self.pop('pore.temperature', None)
        self.pop('pore.pressure', None)

        # Initialize the partition coefficient, K
        self._K = np.ones(self.Nt, dtype=float)
        prefix = self.settings["partition_coef_prefix"]
        self[f"{prefix}.global"] = self._K

        # Add supplied phases to phases dict and initialize occupancy to 0
        self.add_phases(phases)

    def __getitem__(self, key):
        try:
            vals = super().__getitem__(key)
        except KeyError:
            vals = self._interleave_data(key)
        return vals

    @property
    def phases(self):
        phases = {self.project[item].name: self.project[item]
                  for item in self.settings['phases']}
        return phases

    @property
    def K(self):
        self._build_K()
        return self._K

    @K.setter
    def K(self, value):
        self._K = value

    def add_phases(self, phases):
        """
        Adds supplied phases to MultiPhase object and sets occupancy to 0.

        Parameters
        ----------
        phases : list[Phase] or Phase

        """
        phases = np.array(phases, ndmin=1)
        for phase in phases:
            if phase.name in self.settings["phases"]:
                continue
            self.settings['phases'].add(phase.name)
            self[f'pore.occupancy.{phase.name}'] = 0.0
            self[f'throat.occupancy.{phase.name}'] = 0.0

    def set_occupancy(self, phase, *, pores=[], throats=[], values=1):
        r"""
        Specifies occupancy of a phase in each pore or throat. This
        method doesn't return any value.

        Parameters
        ----------
        phase : Phase
            The phase whose occupancy is being specified.
        pores : ndarray
            The location of pores whose occupancy is to be set.
        throats : ndarray
            The location of throats whose occupancy is to be set.
        values : ndarray or float
            Pore/throat occupancy values.

        """
        pores = np.array(pores, ndmin=1)
        throats = np.array(throats, ndmin=1)

        if not(pores.size ^ throats.size):
            raise Exception("Must either pass 'pores' or 'throats'")
        if phase not in self.project:
            raise Exception(f"{phase.name} doesn't belong to this project")
        self.add_phases(phase)

        if pores.size:
            self[f'pore.occupancy.{phase.name}'][pores] = values
        if throats.size:
            self[f'throat.occupancy.{phase.name}'][throats] = values

        if self.settings["throat_occupancy"] == "automatic":
            self.regenerate_models(propnames=f"throat.occupancy.{phase.name}")

    def regenerate_models(self, propnames=None, exclude=[]):
        r"""
        Regenerate models associated with the Multiphase object

        This method works by first regenerating the models associated with
        the constituent phases, and then regenerating Multiphase models.

        Parameters
        ----------
        propnames : list[str] or str
            The list of property names to be regenerated. If None are
            given then ALL models are re-run (except for those whose
            ``regen_mode`` is 'constant').
        exclude : list[str]
            Since the default behavior is to run ALL models, this can be
            used to exclude specific models. It may be more convenient to
            supply as list of 2 models to exclude than to specify 8 models
            to include.

        """
        # Regenerate models associated with phases within MultiPhase object
        for phase in self.phases.values():
            phase.regenerate_models(propnames=propnames, exclude=exclude)
        # Regenerate models specific to MultiPhase object
        super().regenerate_models(propnames=propnames, exclude=exclude)

    def set_binary_partition_coef(self, phases, model, **kwargs):
        """
        Sets binary partition coefficient as defined by the interface
        concentration ratio of phase 1 to phase 2.

        Parameters
        ----------
        phases : list[Phase]
            List of the two phases for which the binary partition
            coefficient model is being added.
        model : OpenPNM model
            Model for calculating the binary partition coefficient.
        kwargs : dict
            Keyword arguments to be passed to the ``model``.

        """
        assert len(phases) == 2
        # Add partition coefficient interface model to the MultiPhase
        propname_prefix = self.settings["partition_coef_prefix"]
        self._add_interface_prop(propname_prefix, phases, model, **kwargs)
        self._build_K()

    def _add_interface_prop(self, propname, phases, model, **kwargs):
        """
        Adds an interface model to the MultiPhase object. See Notes.

        Notes
        -----
        Let's say the two phases corresponding to the interface model are
        named: 'air' and 'water', and the interface propname to be added
        is 'throat.foo'. After augmentation, 'throat.foo.air:water' will
        be the propname that's stored on the ``MultiPhase`` object.

        Note that because of this convention, the order of the phases that
        are passed to this method is important.

        """
        # Add "throat" keyword to the begining of propname if no identifier is found
        if propname.split(".")[0] not in ["pore", "throat"]:
            propname = f"throat.{propname}"
        # Check propname is throat property
        if propname.startswith("pore"):
            raise Exception("'propname' must be a throat property")
        # Add model to Multiphase
        propname = self._format_interface_prop(propname, phases)
        self.add_model(propname, model, **kwargs)

    def _format_interface_prop(self, propname, phases):
        """Formats propname as {propname}.{phase[0].name}:{phase[1].name}"""
        prefix = propname
        suffix = ":".join(phase.name for phase in phases)
        return f"{prefix}.{suffix}"

    def _get_phase_labels(self, formatted_propname):
        """Retrieves phases names from a formatted propname"""
        assert ":" in formatted_propname
        return formatted_propname.split(".")[-1].split(":")

    def _get_interface_throats(self, phase1, phase2):
        conns = self.network.conns
        occ1 = self[f"pore.occupancy.{phase1}"][conns]
        occ2 = self[f"pore.occupancy.{phase2}"][conns]
        idx12, = np.where((occ1[:, 0] == 1) & (occ2[:, 1] == 1))
        idx21, = np.where((occ2[:, 0] == 1) & (occ1[:, 1] == 1))
        return idx12, idx21

    def _build_K(self):
        """Updates the global partition coefficient array"""
        prefix = self.settings["partition_coef_prefix"]
        self._K = np.ones(self.Nt, dtype=float)
        # Find all binary partition coefficient models
        models = [k for k in self.models.keys() if k.startswith(prefix)]
        # Modify the global partition coefficient for each phase pair
        for model in models:
            K12 = self[model]
            phase1, phase2 = self._get_phase_labels(model)
            idx12, idx21 = self._get_interface_throats(phase1, phase2)
            self._K[idx12] = K12[idx12]
            self._K[idx21] = 1 / K12[idx21]
        # Store a reference in self as a propname for convenience
        self[f"{prefix}.global"][:] = self._K

    def _interleave_data(self, prop):
        """Gathers property values from component phases to build a single array."""
        element = self._parse_element(prop)[0]
        vals = np.zeros(self._count(element=element), dtype=float)
        # Retrieve property from constituent phases (weight = occupancy)
        for phase in self.phases.values():
            vals += phase[prop] * self[f"{element}.occupancy.{phase.name}"]
        return vals

    def _set_automatic_throat_occupancy(self, mode="mean"):
        """
        Automatically interpolates throat occupancy based on that in
        adjacent pores. This method doesn't return any value.

        Parameters
        ----------
        mode : str
            Interpolation method. Options are:

            ===========  =====================================================
            mode         meaning
            ===========  =====================================================
            'mean'       sets the throat occupancy as the average of that in
                         adjacent pores.
            'min'        sets the throat occupancy as the minimum value of
                         that in adjacent pores.
            'max'        sets the throat occupancy as the maximum value of
                         that in adjacent pores.
            ===========  =====================================================

        """
        self.settings['throat_occupancy'] = 'automatic'
        for phase in self.phases.values():
            self.add_model(propname=f"throat.occupancy.{phase.name}",
                           model=misc.from_neighbor_pores,
                           prop=f"pore.occupancy.{phase.name}",
                           mode=mode)


def multiphase_diffusion(phase,
                         pore_diffusivity="pore.diffusivity",
                         throat_diffusivity="throat.diffusivity",
                         size_factors="throat.diffusive_size_factors",
                         partition_coef_global="throat.partition_coef.global"):
    r"""
    Calculates the diffusive conductance of conduits for multiphase systems.

    Parameters
    ----------
    %(phase)s
    pore_diffusivity : str
        %(dict_blurb)s pore diffusivity
    throat_diffusivity : str
        %(dict_blurb)s throat diffusivity
    size_factors : str
        %(dict_blurb)s conduit size factors

    Returns
    -------
    %(return_arr)s diffusive conductance

    Notes
    -----
    This method assumes that ``phase["partition_coef"]`` contains information
    on binary phase partitioning. See ``MultiPhase`` class documentation for
    more information.

    """
    network = phase.network
    cn = network.conns
    SF = network[size_factors]
    if isinstance(SF, dict):
        F1, Ft, F2 = SF.values()
    elif SF.ndim > 1:
        F1, Ft, F2 = SF.T
    else:
        F1, Ft, F2 = np.inf, SF, np.inf

    # Fetch model parameters
    D1, D2 = phase[pore_diffusivity][cn].T
    Dt = phase[throat_diffusivity]
    g1 = D1 * F1
    gt = Dt * Ft
    g2 = D2 * F2

    # Apply Henry's partitioning coefficient
    # Note: m12 = (G21*c1 - G12*c2)  NOT  (G12*c1 - G21*c2)
    K12 = phase[partition_coef_global]
    G21 = (1/g1 + 0.5/gt + K12 * (1/g2 + 0.5/gt)) ** -1
    G12 = K12 * G21

    return np.vstack((G12, G21)).T



import numpy as _np
from openpnm.models.geometry import _geodocs
import logging


logger = logging.getLogger(__name__)


__all__ = ["sphere",
           "circle",
           "cube",
           "square"
           ]


@_geodocs
def sphere(
    network,
    pore_diameter='pore.diameter',
    throat_cross_sectional_area='throat.cross_sectional_area',
    amin=0,
):
    r"""
    Calculates internal surface area of pore bodies assuming they are spheres,
    then subtracts the areas of the neighboring throats

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Act)s
    amin : float
        The lower limit for surface area, useful for preventing

    Returns
    -------
    surface_areas : ndarray
        Numpy ndarry containing pore surface area values

    Notes
    -----
    This function subtracts the area of the neighboring throats in a
    crude way, by simply considering the throat cross-sectional area,
    thus not accounting for the actual curvature of the intersection.

    """
    R = network[pore_diameter] / 2
    value = 4 * _np.pi * R**2
    Tca = network[throat_cross_sectional_area]
    _np.subtract.at(value, network.conns[:, 0], Tca)
    _np.subtract.at(value, network.conns[:, 1], Tca)
    if amin is not None:
        value = _np.clip(value, amin, _np.inf)
    elif _np.any(value < 0):
        logger.warn('Negative pore surface areas found')
    return value


@_geodocs
def circle(
    network,
    pore_diameter='pore.diameter',
    throat_cross_sectional_area='throat.cross_sectional_area',
    amin=0,
):
    r"""
    Calculates internal surface area of pore bodies assuming they are
    circular then subtracts the area of the neighboring throats.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Act)s
    amin : float
        The lower limit for surface area, useful for preventing

    Returns
    -------

    Notes
    -----

    """
    value = _np.pi * network[pore_diameter]
    Tca = network[throat_cross_sectional_area]
    _np.subtract.at(value, network.conns[:, 0], Tca)
    _np.subtract.at(value, network.conns[:, 1], Tca)
    if amin is not None:
        value = _np.clip(value, amin, _np.inf)
    elif _np.any(value < 0):
        logger.warn('Negative pore surface areas found')
    return value


def cube(
    network,
    pore_diameter='pore.diameter',
    throat_cross_sectional_area='throat.cross_sectional_area',
    amin=0,
):
    r"""
    Calculates internal surface area of pore bodies assuming they are cubes
    then subtracts the area of the neighboring throats.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Act)s
    amin : float
        The lower limit for surface area, useful for preventing

    Returns
    -------

    """
    D = network[pore_diameter]
    value = 6.0 * D**2
    Tca = network[throat_cross_sectional_area]
    _np.subtract.at(value, network.conns[:, 0], Tca)
    _np.subtract.at(value, network.conns[:, 1], Tca)
    if amin is not None:
        value = _np.clip(value, amin, _np.inf)
    elif _np.any(value < 0):
        logger.warn('Negative pore surface areas found')
    return value


def square(
    network,
    pore_diameter='pore.diameter',
    throat_cross_sectional_area='throat.cross_sectional_area',
    amin=0,
):
    r"""
    Calculates internal surface area of pore bodies assuming they are
    squares then subtracts the area of the neighboring throats.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Act)s
    amin : float
        The lower limit for surface area, useful for preventing

    Returns
    -------

    """
    D = network[pore_diameter]
    value = 4.0 * D
    Tca = network[throat_cross_sectional_area]
    _np.subtract.at(value, network.conns[:, 0], Tca)
    _np.subtract.at(value, network.conns[:, 1], Tca)
    if amin is not None:
        value = _np.clip(value, amin, _np.inf)
    elif _np.any(value < 0):
        logger.warn('Negative pore surface areas found')
    return value



import logging
import numpy as np
import scipy as sp
from scipy.spatial import cKDTree
from scipy.sparse import csgraph
from scipy.spatial import ConvexHull
from openpnm.utils import Workspace
import openpnm._skgraph as skgr


logger = logging.getLogger(__name__)
ws = Workspace()
__all__ = [
    'isoutside',
    'rotate_coords',
    'shear_coords',
    'trim',
    'extend',
    'label_faces',
    'find_surface_pores',
    'dimensionality',
    'clone_pores',
    'merge_networks',
    'stitch',
    'connect_pores',
    'merge_pores',
    'hull_centroid',
    'template_sphere_shell',
    'template_cylinder_annulus',
    'generate_base_points',
    'reflect_base_points',
    'add_boundary_pores',
    'iscoplanar',
    'is_fully_connected',
    'get_spacing',
    'get_shape',
    'get_domain_area',
    'get_domain_length',
    'filter_pores_by_z',
    'find_interface_throats',
    'add_reservoir_pore',
    'reduce_coordination',
]


def isoutside(**kwargs):
    return skgr.tools.isoutside(**kwargs)


isoutside.__doc__ = skgr.tools.isoutside.__doc__


def rotate_coords(network, **kwargs):
    network['pore.coords'] = skgr.tools.rotate_coords(network.coords, **kwargs)
    return network


rotate_coords.__doc__ = skgr.tools.rotate_coords.__doc__


def shear_coords(network, **kwargs):
    network['pore.coords'] = skgr.tools.shear_coords(network.coords, **kwargs)
    return network


shear_coords.__doc__ = skgr.tools.shear_coords.__doc__


def template_sphere_shell(r_outer, r_inner=0):
    return skgr.generators.tools.template_sphere_shell(r_outer, r_inner)


template_sphere_shell.__doc__ = \
    skgr.generators.tools.template_sphere_shell.__doc__


def template_cylinder_annulus(z, r_outer, r_inner=0):
    return skgr.generators.tools.template_cylinder_annulus(z, r_outer, r_inner)


template_cylinder_annulus.__doc__ = \
    skgr.generators.tools.template_cylinder_annulus.__doc__


def generate_base_points(num_points, domain_size, reflect=True):
    return skgr.generators.tools.generate_base_points(
        num_points,
        domain_size,
        reflect,
    )


generate_base_points.__doc__ = \
    skgr.generators.tools.generate_base_points.__doc__


def reflect_base_points(points, domain_size):
    return skgr.generators.tools.reflect_base_points(points, domain_size)


reflect_base_points.__doc__ = \
    skgr.generators.tools.reflect_base_points.__doc__


def get_spacing(network):
    return skgr.tools.get_cubic_spacing(network)


get_spacing.__doc__ = skgr.tools.get_cubic_spacing.__doc__


def get_shape(network):
    return skgr.tools.get_cubic_shape(network)


get_shape.__doc__ = skgr.tools.get_cubic_shape.__doc__


def filter_pores_by_z(network, pores, z):
    return skgr.queries.filter_by_z(network=network, inds=pores, z=z)


filter_pores_by_z.__doc__ = skgr.queries.filter_by_z.__doc__


def find_interface_throats(network, P1, P2):
    return skgr.queries.find_common_edges(network=network, inds_1=P1, inds_2=P2)


find_interface_throats.__doc__ = skgr.queries.find_common_edges.__doc__


def dimensionality(network):
    r"""
    Determines whether a network is 1D, 2D or 3D, and in which dimensions

    Parameters
    ----------
    network : dict
        The OpenPNM network of interest

    Returns
    -------
    dims : boolean mask
        A 3 x 1 array of booleans with ``True`` indicating if any
        dimensionality exists on each axis.

    Notes
    -----
    This function looks at the coordinates of each pore and if any axes all
    have the same values that axis is considered non-dimensional.

    """
    return skgr.tools.dimensionality(network)


def trim(network, pores=[], throats=[]):
    """
    Remove pores or throats from the network

    Parameters
    ----------
    network : Network
        The Network from which pores or throats should be removed
    pores (or throats) : array_like
        The indices of the of the pores or throats to be removed from the
        network.

    """
    pores = network._parse_indices(pores)
    throats = network._parse_indices(throats)
    Pkeep = np.copy(network['pore.all'])
    Tkeep = np.copy(network['throat.all'])
    if np.size(pores) > 0:
        Pkeep[pores] = False
        if not np.any(Pkeep):
            raise Exception('Cannot delete ALL pores')
        # # Performing customized find_neighbor_throats which is much faster,
        # # but not general for other types of queries
        # temp = np.in1d(network['throat.conns'].flatten(), pores)
        # temp = np.reshape(temp, (network.Nt, 2))
        # Ts = np.any(temp, axis=1)
        # Ts = network.Ts[Ts]
        Ts = network.find_neighbor_throats(pores=~Pkeep, mode='union')
        if len(Ts) > 0:
            Tkeep[Ts] = False
    if np.size(throats) > 0:
        Tkeep[throats] = False
        # The following IF catches the special case of deleting ALL throats
        # It removes all throat props, adds 'all', and skips rest of function
        if not np.any(Tkeep):
            logger.info('Removing ALL throats from network')
            for item in list(network.keys()):
                if item.split('.', 1)[0] == 'throat':
                    del network[item]
            network['throat.all'] = np.array([], ndmin=1)
            return

    # Temporarily store throat conns and pore map for processing later
    Np_old = network.Np
    Nt_old = network.Nt
    Pkeep_inds = np.where(Pkeep)[0]
    Tkeep_inds = np.where(Tkeep)[0]
    Pmap = np.ones((network.Np,), dtype=int)*-1
    tpore1 = network['throat.conns'][:, 0]
    tpore2 = network['throat.conns'][:, 1]

    # Delete specified pores and throats from all objects
    for obj in network.project:
        if (obj.Np == Np_old) and (obj.Nt == Nt_old):
            Ps = Pkeep_inds
            Ts = Tkeep_inds
        for key in list(obj.keys()):
            temp = obj.pop(key)
            if key.split('.', 1)[0] == 'throat':
                obj.update({key: temp[Ts]})
            if key.split('.', 1)[0] == 'pore':
                obj.update({key: temp[Ps]})

    # Remap throat connections
    Pmap[Pkeep] = np.arange(0, np.sum(Pkeep))
    Tnew1 = Pmap[tpore1[Tkeep]]
    Tnew2 = Pmap[tpore2[Tkeep]]
    network.update({'throat.conns': np.vstack((Tnew1, Tnew2)).T})

    # Clear adjacency and incidence matrices which will be out of date now
    network._am.clear()
    network._im.clear()


def extend(network, coords=[], conns=[], labels=[], **kwargs):
    r"""
    Add pores or throats to the network from a list of coords or conns.

    Parameters
    ----------
    network : Network
        The network to which pores or throats should be added
    coords : array_like
        The coordinates of the pores to add.  These will be appended to the
        'pore.coords' array so should be of shape N-by-3, where N is the
        number of pores in the list.
    conns : array_like
        The throat connections to add.  These will be appended to the
        'throat.conns' array so should be of shape N-by-2.  Note that the
        numbering must point to existing pores.
    labels : str, or list[str], optional
        A list of labels to apply to the new pores and throats

    """
    if 'throat_conns' in kwargs.keys():
        conns = kwargs['throat_conns']
    if 'pore_coords' in kwargs.keys():
        coords = kwargs['pore_coords']
    coords = np.array(coords)
    conns = np.array(conns)
    Np_old = network.num_pores()
    Nt_old = network.num_throats()
    Np = Np_old + coords.shape[0]
    Nt = Nt_old + conns.shape[0]
    if np.any(conns > Np):
        raise Exception('Some throat conns point to non-existent pores')
    network.update({'pore.all': np.ones([Np, ], dtype=bool),
                    'throat.all': np.ones([Nt, ], dtype=bool)})
    # Add coords and conns
    if np.size(coords) > 0:
        coords = np.vstack((network['pore.coords'], coords))
        network['pore.coords'] = coords
    if np.size(conns) > 0:
        conns = np.vstack((network['throat.conns'], conns))
        network['throat.conns'] = conns

    # Increase size of any prop or label arrays already on network and phases
    objs = list(network.project.phases)
    objs.append(network)
    for obj in objs:
        obj.update({'pore.all': np.ones([Np, ], dtype=bool),
                    'throat.all': np.ones([Nt, ], dtype=bool)})
        for item in list(obj.keys()):
            N = obj._count(element=item.split('.', 1)[0])
            if obj[item].shape[0] < N:
                arr = obj.pop(item)
                s = arr.shape
                if arr.dtype == bool:
                    obj[item] = np.zeros(shape=(N, *s[1:]), dtype=bool)
                else:
                    obj[item] = np.ones(shape=(N, *s[1:]), dtype=float)*np.nan
                obj[item][:arr.shape[0]] = arr

    # Regenerate models on all objects to fill new elements
    for obj in network.project.phases:
        if hasattr(obj, 'models'):
            obj.regenerate_models()

    # Apply labels, if supplied
    if labels != []:
        # Convert labels to list if necessary
        if isinstance(labels, str):
            labels = [labels]
        for label in labels:
            # Remove pore or throat from label, if present
            label = label.split('.', 1)[-1]
            if np.size(coords) > 0:
                Ps = np.r_[Np_old:Np]
                if 'pore.'+label not in network.labels():
                    network['pore.'+label] = False
                network['pore.'+label][Ps] = True
            if np.size(conns) > 0:
                Ts = np.r_[Nt_old:Nt]
                if 'throat.'+label not in network.labels():
                    network['throat.'+label] = False
                network['throat.'+label][Ts] = True

    # Clear adjacency and incidence matrices which will be out of date now
    network._am.clear()
    network._im.clear()


def label_faces(network, tol=0.0, label='surface'):
    r"""
    Finds pores on the surface of the network and labels them according to
    whether they are on the *top*, *bottom*, etc.

    This function assumes the network is cubic in shape

    Parameters
    ----------
    network : Network
        The network to apply the labels
    tol : scalar
        The tolerance for defining what counts as a surface pore, which is
        specifically meant for random networks.  All pores with ``tol`` of
        the maximum or minimum along each axis are counts as pores.  The
        default is 0.
    label : str
        An identifying label to isolate the pores on the faces of the network.
        The default is 'surface'.  Surface pores can be found using
        ``find_surface_pores``.

    """
    if label is not None:
        label = label.split('.', 1)[-1]
        if 'pore.'+label not in network.labels():
            find_surface_pores(network, label=label)
        Psurf = network['pore.'+label]
    else:
        Psurf = True  # So it will "do nothing" below
    crds = network['pore.coords']
    xmin, xmax = np.amin(crds[:, 0]), np.amax(crds[:, 0])
    xspan = xmax - xmin
    ymin, ymax = np.amin(crds[:, 1]), np.amax(crds[:, 1])
    yspan = ymax - ymin
    zmin, zmax = np.amin(crds[:, 2]), np.amax(crds[:, 2])
    zspan = zmax - zmin
    dims = dimensionality(network)
    if dims[0]:
        network['pore.left'] = (crds[:, 0] <= (xmin + tol*xspan)) * Psurf
        network['pore.xmin'] = np.copy(network['pore.left'])
        network['pore.right'] = (crds[:, 0] >= (xmax - tol*xspan)) * Psurf
        network['pore.xmax'] = np.copy(network['pore.right'])
    if dims[1]:
        network['pore.front'] = (crds[:, 1] <= (ymin + tol*yspan)) * Psurf
        network['pore.ymin'] = np.copy(network['pore.front'])
        network['pore.back'] = (crds[:, 1] >= (ymax - tol*yspan)) * Psurf
        network['pore.ymax'] = np.copy(network['pore.back'])
    if dims[2]:
        network['pore.bottom'] = (crds[:, 2] <= (zmin + tol*zspan)) * Psurf
        network['pore.zmin'] = np.copy(network['pore.bottom'])
        network['pore.top'] = (crds[:, 2] >= (zmax - tol*zspan)) * Psurf
        network['pore.zmax'] = np.copy(network['pore.top'])


def find_surface_pores(network, markers=None, label='surface'):
    r"""
    Find the pores on the surface of the domain by performing a Delaunay
    triangulation between the network pores and some external ``markers``. All
    pores connected to these external marker points are considered surface
    pores.

    Parameters
    ----------
    network: Network
        The network for which the surface pores are to be found
    markers: array_like
        3 x N array of the marker coordinates to use in the triangulation.  The
        labeling is performed in one step, so all points are added, and then
        any pores connected to at least one marker is given the provided label.
        By default, this function will automatically generate 6 points outside
        each axis of the network domain. Users may wish to specify a single
        external marker point and provide an appropriate label in order to
        identify specific faces.  For instance, the marker may be *above* the
        domain, and the label might be 'top_surface'.
    label : str
        The label to apply to the pores.  The default is 'surface'.

    Notes
    -----
    This function does not check whether the given markers actually lie outside
    the domain, allowing the labeling of *internal* sufaces.

    If this method fails to mark some surface pores, consider sending more
    markers on each face.

    """
    import scipy.spatial as sptl
    dims = dimensionality(network)
    coords = network['pore.coords'][:, dims]
    if markers is None:
        # normalize coords to a 1 unit cube centered on origin
        coords -= np.amin(coords, axis=0)
        coords /= np.amax(coords, axis=0)
        coords -= 0.5
        npts = max((network.Np/10, 100))
        if sum(dims) == 1:
            network['pore.'+label] = True
            return
        if sum(dims) == 2:
            r = 0.75
            theta = np.linspace(0, 2*np.pi, int(npts), dtype=float)
            x = r*np.cos(theta)
            y = r*np.sin(theta)
            markers = np.vstack((x, y)).T
        if sum(dims) == 3:
            r = 1.00
            indices = np.arange(0, int(npts), dtype=float) + 0.5
            phi = np.arccos(1 - 2*indices/npts)
            theta = np.pi * (1 + 5**0.5) * indices
            x = r*np.cos(theta) * np.sin(phi)
            y = r*np.sin(theta) * np.sin(phi)
            z = r*np.cos(phi)
            markers = np.vstack((x, y, z)).T
    else:
        if sum(dims) == 1:
            pass
        if sum(dims) == 2:
            markers = np.atleast_2d(markers)
            if markers.shape[1] != 2:
                raise Exception('Network appears planar, so markers must be 2D')
        if sum(dims) == 3:
            markers = np.atleast_2d(markers)
            if markers.shape[1] != 3:
                raise Exception('Markers must be 3D for this network')
    pts = np.vstack((coords, markers))
    tri = sptl.Delaunay(pts, incremental=False)
    (indices, indptr) = tri.vertex_neighbor_vertices
    for k in range(network.Np, tri.npoints):
        neighbors = indptr[indices[k]:indices[k+1]]
        inds = np.where(neighbors < network.Np)
        neighbors = neighbors[inds]
        if 'pore.'+label not in network.keys():
            network['pore.'+label] = False
        network['pore.'+label][neighbors] = True


def clone_pores(network, pores, labels=['clone'], mode='parents'):
    r"""
    Clones the specified pores and adds them to the network

    Parameters
    ----------
    network : Network
        The Network object to which the new pores are to be added
    pores : array_like
        List of pores to clone
    labels : str, or list[str]
        The labels to apply to the clones, default is 'clone'
    mode : str
        Controls the connections between parents and clones.  Options are:

        ===========  ==========================================================
        mode         description
        ===========  ==========================================================
        'parents'    Each clone is connected only to its parent.(Default)
        'siblings'   Clones are only connected to each other in the same
                     manner as parents were connected.
        'isolated'   No connections between parents or siblings
        ===========  ==========================================================

    """
    if isinstance(labels, str):
        labels = [labels]
    network._parse_indices(pores)
    Np = network.Np
    Nt = network.Nt
    # Clone pores
    parents = np.array(pores, ndmin=1)
    pcurrent = network['pore.coords']
    pclone = pcurrent[pores, :]
    pnew = np.concatenate((pcurrent, pclone), axis=0)
    Npnew = np.shape(pnew)[0]
    clones = np.arange(Np, Npnew)
    # Create cloned pores first
    extend(network=network, pore_coords=pclone)
    # Apply provided labels to cloned pores
    for item in labels:
        network.set_label(label=item, pores=range(Np, Npnew))
    # Add connections between parents and clones
    if mode == 'parents':
        tclone = np.vstack((parents, clones)).T
        extend(network=network, conns=tclone)
    elif mode == 'siblings':
        ts = network.find_neighbor_throats(pores=pores, mode='xnor')
        mapping = np.zeros([network.Np, ], dtype=int)
        mapping[pores] = np.arange(Np, network.Np)
        tclone = mapping[network['throat.conns'][ts]]
        extend(network=network, throat_conns=tclone)
    elif mode == 'isolated':
        pass
    Ntnew = network.Nt
    for item in labels:
        network.set_label(label=item, throats=range(Nt, Ntnew))

    # Clear adjacency and incidence matrices which will be out of date now
    network._am.clear()
    network._im.clear()


def merge_networks(network, donor=[]):
    r"""
    Combine multiple networks into one without making any topological
    connections

    Parameters
    ----------
    network : Network
        The network to which all the other networks should be added.
    donor : Network or list of Objects
        The network object(s) to add to the given network

    See Also
    --------
    extend
    trim
    stitch

    """
    if isinstance(donor, list):
        donors = donor
    else:
        donors = [donor]

    for donor in donors:
        network['throat.conns'] = np.vstack((network['throat.conns'],
                                             donor['throat.conns']
                                             + network.Np))
        network['pore.coords'] = np.vstack((network['pore.coords'],
                                            donor['pore.coords']))
        p_all = np.ones((np.shape(network['pore.coords'])[0],), dtype=bool)
        t_all = np.ones((np.shape(network['throat.conns'])[0],), dtype=bool)
        network.update({'pore.all': p_all})
        network.update({'throat.all': t_all})
        for key in set(network.keys()).union(set(donor.keys())):
            if key.split('.')[1] not in ['conns', 'coords', '_id', 'all']:
                if key in network.keys():
                    pop_flag = False
                    # If key not on donor add it first with dummy values to
                    # simplify merging later
                    if key not in donor.keys():
                        logger.debug('Adding ' + key + ' to donor')
                        if network[key].dtype == bool:  # Deal with labels
                            donor[key] = False
                        else:  # Deal with numerical data
                            element = key.split('.', 1)[0]
                            shape = list(network[key].shape)
                            N = donor._count(element)
                            shape[0] = N
                            donor[key] = np.empty(shape=shape)*np.nan
                        pop_flag = True
                    # Then merge it with existing array on network
                    if len(network[key].shape) == 1:
                        temp = np.hstack((network[key], donor[key]))
                    else:
                        temp = np.vstack((network[key], donor[key]))
                    network[key] = temp
                    if pop_flag:
                        donor.pop(key, None)
                else:
                    # If key not on network add it first
                    logger.debug('Adding ' + key + ' to network')
                    if donor[key].dtype == bool:
                        network[key] = False
                    else:
                        data_shape = list(donor[key].shape)
                        pore_prop = True if key.split(".")[0] == "pore" else False
                        data_shape[0] = network.Np if pore_prop else network.Nt
                        network[key] = np.empty(data_shape) * np.nan
                    # Then append donor values to network
                    s = np.shape(donor[key])[0]
                    network[key][-s:] = donor[key]

    # Clear adjacency and incidence matrices which will be out of date now
    network._am.clear()
    network._im.clear()


def stitch(network, donor, P_network, P_donor, method='nearest',
           len_max=np.inf, label_suffix='', label_stitches='stitched'):
    r"""
    Stitches a second a network to the current network.

    Parameters
    ----------
    network : Network
        The Network to which to donor Network will be attached
    donor : Network
        The Network to stitch on to the current Network
    P_network : array_like
        The pores on the current Network
    P_donor : array_like
        The pores on the donor Network
    label_suffix : str or None
        Some text to append to each label in the donor Network before
        inserting them into the recipient.  The default is to append no
        text, but a common option would be to append the donor Network's
        name. To insert none of the donor labels, use ``None``.
    label_stitches : str or list[str]
        The label to apply to the newly created 'stitch' throats.  The
        defaul is 'stitched'.  If performing multiple stitches in a row it
        might be helpful to the throats created during each step uniquely
        for later identification.
    len_max : float
        Set a length limit on length of new throats
    method : str (default = 'nearest')
        The method to use when making pore to pore connections. Options are:

        ========= =============================================================
        mode      description
        ========= =============================================================
        'radius'  Connects each pore on the recipient network to the nearest
                  pores on the donor network, within ``len_max``
        'nearest' Connects each pore on the recipienet network to the nearest
                  pore on the donor network.
        ========= =============================================================

    Notes
    -----
    Before stitching it is necessary to translate the pore coordinates of
    one of the Networks so that it is positioned correctly relative to the
    other.  This is illustrated in the example below.

    """
    # Parse inputs
    if isinstance(label_stitches, str):
        label_stitches = [label_stitches]
    for s in label_stitches:
        if s not in network.keys():
            network['throat.' + s] = False
    # Get the initial number of pores and throats
    N_init = {}
    N_init['pore'] = network.Np
    N_init['throat'] = network.Nt
    if method == 'nearest':
        P1 = P_network
        P2 = P_donor + N_init['pore']  # Increment pores on donor
        C1 = network['pore.coords'][P_network]
        C2 = donor['pore.coords'][P_donor]
        D = sp.spatial.distance.cdist(C1, C2)
        [P1_ind, P2_ind] = np.where(D == D.min(axis=0))
        conns = np.vstack((P1[P1_ind], P2[P2_ind])).T
    elif method == 'radius':
        P1 = P_network
        P2 = P_donor + N_init['pore']  # Increment pores on donor
        C1 = network['pore.coords'][P_network]
        C2 = donor['pore.coords'][P_donor]
        D = sp.spatial.distance.cdist(C1, C2)
        [P1_ind, P2_ind] = np.where(D <= len_max)
        conns = np.vstack((P1[P1_ind], P2[P2_ind])).T
    else:
        raise Exception('<{}> method not supported'.format(method))

    merge_networks(network, donor)

    # Add the new stitch throats to the Network
    extend(network=network, throat_conns=conns, labels=label_stitches)

    # Remove donor from Workspace, if present
    # This check allows for the reuse of a donor Network multiple times
    for sim in list(ws.values()):
        if donor in sim:
            del ws[sim.name]


def stitch_pores(network, pores1, pores2, mode='gabriel'):
    r"""
    Stitches together pores in a network with disconnected clusters

    Parameters
    ----------
    network : OpenPNM Network
        The network to operate upon
    pores1 and pores2: array_like
        The pore indices of the disconnected clusters to be joined
    mode : str
        Dictates which tesselation method is used to identify which pores to
        stitch together.  Options are:

        ===========  ==========================================================
        mode         meaning
        ===========  ==========================================================
        'delaunay'   Uses the delaunay tesselation method
        ===========  ==========================================================

    Returns
    -------
    None
        The network is operated on 'in-place' so nothing is returned.

    """
    raise NotImplementedError()
    from openpnm.network import Delaunay
    pores1 = network._parse_indices(pores1)
    pores2 = network._parse_indices(pores2)
    C1 = network.coords[pores1, :]
    C2 = network.coords[pores2, :]
    crds = np.vstack((C1, C2))
    if mode == 'delaunay':
        shape = get_shape(network) * dimensionality(network)
        net = Delaunay(points=crds, shape=shape)
    else:
        raise Exception('Unsupported mode')
    net.set_label(pores=range(len(pores1)), label='pore.one')
    net.set_label(pores=range(len(pores2)), label='pore.two')
    Ts = net.find_neighbor_throats(pores=net.pores('one'), mode='xor')
    conns = net.conns[Ts]
    mapped_conns = np.vstack((pores1[conns[:, 0]],
                              pores2[conns[:, 1] - len(pores1)])).T
    mapped_conns = np.sort(mapped_conns, axis=1)
    extend(network=network, conns=mapped_conns, labels='stitched')


def connect_pores(network, pores1, pores2, labels=['new_conns']):
    r"""
    Returns the possible connections between two groups of pores

    Parameters
    ----------
    network : Network
        The network to which the pores should be added
    pores1 : array_like
        The first group of pores on the network
    pores2 : array_like
        The second group of pores on the network
    labels : list of strings
        The labels to apply to the new throats. The default is ``'new_conns'``.

    Notes
    -----
    The method also works if ``pores1`` and ``pores2`` are list of lists,
    in which case it consecutively connects corresponding members of the two
    lists in a 1-to-1 fashion. Example: pores1 = [[0, 1], [2, 3]] and
    pores2 = [[5], [7, 9]] leads to creation of the following connections:

    ::

        0 --> 5     2 --> 7     3 --> 7
        1 --> 5     2 --> 9     3 --> 9

    If you want to use the batch functionality, make sure that each element
    within ``pores1`` and ``pores2`` are of type list or ndarray.

    """
    # Assert that `pores1` and `pores2` are list of lists
    try:
        len(pores1[0])
    except (TypeError, IndexError):
        pores1 = [pores1]
    try:
        len(pores2[0])
    except (TypeError, IndexError):
        pores2 = [pores2]

    if len(pores1) != len(pores2):
        raise Exception('Running in batch mode! pores1 and pores2 must be'
                        + ' of the same length.')

    arr1, arr2 = [], []
    for ps1, ps2 in zip(pores1, pores2):
        size1 = np.size(ps1)
        size2 = np.size(ps2)
        arr1.append(np.repeat(ps1, size2))
        arr2.append(np.tile(ps2, size1))
    conns = np.vstack([np.concatenate(arr1), np.concatenate(arr2)]).T
    extend(network=network, throat_conns=conns, labels=labels)


def merge_pores(network, pores, labels=['merged'], include_neighbors=True):
    r"""
    Combines a selection of pores into a new single pore located at the
    centroid of the selected pores (and optionally their neighbors)
    and connected to all of their neighbors.

    Parameters
    ----------
    network : Network
    pores : array_like
        The list of pores which are to be combined into a new single pore
    labels : str or list[str]
        The labels to apply to the new pore and new throat connections

    Notes
    -----
    (1) The method also works if a list of lists is passed, in which case
    it consecutively merges the given selections of pores.

    (2) The selection of pores should be chosen carefully, preferrable so that
    they all form a continuous cluster.  For instance, it is recommended
    to use the ``find_nearby_pores`` method to find all pores within a
    certain distance of a given pore, and these can then be merged without
    causing any abnormal connections.

    """
    # Assert that `pores` is list of lists
    try:
        len(pores[0])
    except (TypeError, IndexError):
        pores = [pores]

    N = len(pores)
    NBs, XYZs = [], []

    for Ps in pores:
        temp = network.find_neighbor_pores(pores=Ps,
                                           mode='union',
                                           flatten=True,
                                           include_input=False)
        NBs.append(temp)
        if len(Ps) == 2:
            XYZs.append(np.mean(network["pore.coords"][Ps], axis=0))
        else:
            if include_neighbors:
                points = np.concatenate((temp, Ps))
            else:
                points = Ps
            XYZs.append(hull_centroid(network["pore.coords"][points]))

    extend(network, pore_coords=XYZs, labels=labels)
    Pnew = network.Ps[-N::]

    # Possible throats between new pores: This only happens when running in
    # batch mode, i.e. multiple groups of pores are to be merged. In case
    # some of these groups share elements, possible throats between the
    # intersecting elements is not captured and must be added manually.
    pores_set = [set(items) for items in pores]
    NBs_set = [set(items) for items in NBs]
    ps1, ps2 = [], []
    from itertools import combinations
    for i, j in combinations(range(N), 2):
        if not NBs_set[i].isdisjoint(pores_set[j]):
            ps1.append([network.Ps[-N+i]])
            ps2.append([network.Ps[-N+j]])

    # Add (possible) connections between the new pores
    connect_pores(network, pores1=ps1, pores2=ps2, labels=labels)
    # Add connections between the new pores and the rest of the network
    connect_pores(network, pores2=np.split(Pnew, N), pores1=NBs, labels=labels)
    # Trim merged pores from the network
    trim(network=network, pores=np.concatenate(pores))


def hull_centroid(points):
    r"""
    Computes centroid of the convex hull enclosing the given coordinates.

    Parameters
    ----------
    points : Np by 3 ndarray
        Coordinates (xyz)

    Returns
    -------
    centroid : array
        A 3 by 1 Numpy array containing coordinates of the centroid.

    """
    dim = [np.unique(points[:, i]).size != 1 for i in range(3)]
    hull = ConvexHull(points[:, dim])
    centroid = points.mean(axis=0)
    centroid[dim] = hull.points[hull.vertices].mean(axis=0)

    return centroid


def add_boundary_pores(network, pores, offset=None, move_to=None,
                       apply_label='boundary'):
    r"""
    This method uses ``clone_pores`` to clone the input pores, then shifts
    them the specified amount and direction, then applies the given label.

    Parameters
    ----------
    pores : array_like
        List of pores to offset.  If no pores are specified, then it
        assumes that all surface pores are to be cloned.
    offset : 3 x 1 array
        The distance in vector form which the cloned boundary pores should
        be offset.  Either this, or ``move_to`` must be specified.
    move_to : 3 x 1 array
        The location to move the boundary pores to.  A value of ``None``
        indicates that no translation should be applied in that axis.  For
        instance, ``[None, None, 0]`` indicates that the boundary pores should
        moved along the z-axis to the specified location.  Either this or
        ``offset`` must be specified.
    apply_label : str
        This label is applied to the boundary pores.  Default is
        'boundary'.

    """
    # Parse the input pores
    Ps = np.array(pores, ndmin=1)
    if Ps.dtype is bool:
        Ps = network.to_indices(Ps)
    if np.size(pores) == 0:  # Handle an empty array if given
        return np.array([], dtype=np.int64)
    # Clone the specifed pores
    clone_pores(network=network, pores=Ps)
    newPs = network.pores('pore.clone')
    del network['pore.clone']
    newTs = network.throats('clone')
    del network['throat.clone']
    if offset is not None:  # Offset the cloned pores
        network['pore.coords'][newPs] += offset
    if move_to is not None:  # Move the cloned pores
        for i, d in enumerate(move_to):
            if d is not None:
                temp = network['pore.coords'][newPs]
                temp[:, i] = d
                network['pore.coords'][newPs] = temp
    # Apply labels to boundary pores (trim leading 'pores' if present)
    label = apply_label.split('.', 1)[-1]
    plabel = 'pore.' + label
    tlabel = 'throat.' + label
    network[plabel] = False
    network[plabel][newPs] = True
    network[tlabel] = False
    network[tlabel][newTs] = True


def iscoplanar(coords):
    r"""
    Determines if given pores are coplanar with each other

    Parameters
    ----------
    coords : array_like
        List of pore coords to check for coplanarity.  At least 3 pores are
        required.

    Returns
    -------
    results : bool
        A boolean value of whether given points are coplanar (``True``) or
        not (``False``)

    """
    coords = np.array(coords, ndmin=1)
    if np.shape(coords)[0] < 3:
        raise Exception('At least 3 input pores are required')

    Px = coords[:, 0]
    Py = coords[:, 1]
    Pz = coords[:, 2]

    # Do easy check first, for common coordinate
    if np.shape(np.unique(Px))[0] == 1:
        return True
    if np.shape(np.unique(Py))[0] == 1:
        return True
    if np.shape(np.unique(Pz))[0] == 1:
        return True

    # Perform rigorous check using vector algebra
    # Grab first basis vector from list of coords
    n1 = np.array((Px[1] - Px[0], Py[1] - Py[0], Pz[1] - Pz[0])).T
    n = np.array([0.0, 0.0, 0.0])
    i = 1
    while n.sum() == 0:
        if i >= (np.size(Px) - 1):
            logger.warning('No valid basis vectors found')
            return False
        # Chose a secon basis vector
        n2 = np.array((Px[i+1] - Px[i], Py[i+1] - Py[i], Pz[i+1] - Pz[i])).T
        # Find their cross product
        n = np.cross(n1, n2)
        i += 1
    # Create vectors between all other pairs of points
    r = np.array((Px[1:-1] - Px[0], Py[1:-1] - Py[0], Pz[1:-1] - Pz[0]))
    # Ensure they all lie on the same plane
    n_dot = np.dot(n, r)

    return bool(np.sum(np.absolute(n_dot)) == 0)


def is_fully_connected(network, pores_BC=None):
    r"""
    Checks whether network is fully connected, i.e. not clustered.

    Parameters
    ----------
    network : Network
        The network whose connectivity to check.
    pores_BC : array_like (optional)
        The pore indices of boundary conditions (inlets/outlets).

    Returns
    -------
    bool
        If ``pores_BC`` is not specified, then returns ``True`` only if
        the entire network is connected to the same cluster. If
        ``pores_BC`` is given, then returns ``True`` only if all clusters
        are connected to the given boundary condition pores.

    """
    am = network.get_adjacency_matrix(fmt='lil').copy()
    temp = csgraph.connected_components(am, directed=False)[1]
    is_connected = np.unique(temp).size == 1
    # Ensure all clusters are part of pores, if given
    if not is_connected and pores_BC is not None:
        am.resize(network.Np + 1, network.Np + 1)
        pores_BC = network._parse_indices(pores_BC)
        am.rows[-1] = pores_BC.tolist()
        am.data[-1] = np.arange(network.Nt, network.Nt + len(pores_BC)).tolist()
        temp = csgraph.connected_components(am, directed=False)[1]
        is_connected = np.unique(temp).size == 1
    return is_connected


def get_domain_area(network, inlets=None, outlets=None):
    r"""
    Determine the cross sectional area relative to the inlets/outlets.

    Parameters
    ----------
    network : Network
        The network object containing the pore coordinates
    inlets : array_like
        The pore indices of the inlets.
    outlets : array_Like
        The pore indices of the outlets.

    Returns
    -------
    area : scalar
        The cross sectional area relative to the inlets/outlets.

    """
    logger.warning('Attempting to estimate inlet area...will be low')
    if dimensionality(network).sum() != 3:
        raise Exception('The network is not 3D, specify area manually')
    inlets = network.coords[inlets]
    outlets = network.coords[outlets]
    if not iscoplanar(inlets):
        logger.error('Detected inlet pores are not coplanar')
    if not iscoplanar(outlets):
        logger.error('Detected outlet pores are not coplanar')
    Nin = np.ptp(inlets, axis=0) > 0
    if Nin.all():
        logger.warning('Detected inlets are not oriented along a principle axis')
    Nout = np.ptp(outlets, axis=0) > 0
    if Nout.all():
        logger.warning('Detected outlets are not oriented along a principle axis')
    hull_in = ConvexHull(points=inlets[:, Nin])
    hull_out = ConvexHull(points=outlets[:, Nout])
    if hull_in.volume != hull_out.volume:
        logger.error('Inlet and outlet faces are different area')
    area = hull_in.volume  # In 2D: volume=area, area=perimeter
    return area


def get_domain_length(network, inlets=None, outlets=None):
    r"""
    Determine the domain length relative to the inlets/outlets.

    Parameters
    ----------
    network : Network
        The network object containing the pore coordinates
    inlets : array_like
        The pore indices of the inlets.
    outlets : array_Like
        The pore indices of the outlets.

    Returns
    -------
    area : scalar
        The domain length relative to the inlets/outlets.

    """
    msg = ('Attempting to estimate domain length...could be low if'
           ' boundary pores were not added')
    logger.warning(msg)
    inlets = network.coords[inlets]
    outlets = network.coords[outlets]
    if not iscoplanar(inlets):
        logger.error('Detected inlet pores are not coplanar')
    if not iscoplanar(outlets):
        logger.error('Detected inlet pores are not coplanar')
    tree = cKDTree(data=inlets)
    Ls = np.unique(np.float64(tree.query(x=outlets)[0]))
    if not np.allclose(Ls, Ls[0]):
        logger.error('A unique value of length could not be found')
    length = Ls[0]
    return length


def reduce_coordination(network, z):
    r"""
    Deletes throats on network to match specified average coordination number

    Parameters
    ----------
    target : Network
        The network whose throats are to be trimmed
    z : scalar
        The desired average coordination number.  It is not possible to specify
        the distribution of the coordination, only the mean value.

    Returns
    -------
    trim : ndarray
        A boolean array with ``True`` values indicating which pores to trim
        (using ``op.topotools.trim``) to obtain the desired average
        coordination number.

    Notes
    -----
    This method first finds the minimum spanning tree of the network using
    random weights on each throat, then assures that these throats are *not*
    deleted, in order to maintain network connectivity.  The list of throats
    to trim is generated randomly from the throats *not* on the spanning tree.

    """
    # Find minimum spanning tree using random weights
    am = network.create_adjacency_matrix(weights=np.random.rand(network.Nt),
                                         triu=False)
    mst = csgraph.minimum_spanning_tree(am, overwrite=True)
    mst = mst.tocoo()

    # Label throats on spanning tree to avoid deleting them
    Ts = network.find_connecting_throat(mst.row, mst.col)
    Ts = np.hstack(Ts)
    network['throat.mst'] = False
    network['throat.mst'][Ts] = True

    # Trim throats not on the spanning tree to acheive desired coordination
    Ts = np.random.permutation(network.throats('mst', mode='nor'))
    del network['throat.mst']
    Ts = Ts[:int(network.Nt - network.Np*(z/2))]
    Ts = network.to_mask(throats=Ts)
    return Ts


def add_reservoir_pore(cls, network, pores, offset=0.1):
    r"""
    Adds a single pore connected to all ``pores`` to act as a reservoir

    This function is mostly needed to make network compatible with the
    Statoil file format, which requires reservoir pores on the inlet and
    outlet faces.

    Parameters
    ----------
    network : Network
        The network to which the reservoir pore should be added
    pores : array_like
        The pores to which the reservoir pore should be connected to
    offset : scalar
        Controls the distance which the reservoir is offset from the given
        ``pores``.  The total displacement is found from the network
        dimension normal to given ``pores``, multiplied by ``offset``.

    Returns
    -------
    project : list
        An OpenPNM project object with a new geometry object added to
        represent the newly added pore and throats.

    """
    import openpnm.models.geometry as mods
    # Check if a label was given and fetch actual indices
    if isinstance(pores, str):
        # Convert 'face' into 'pore.face' if necessary
        if not pores.startswith('pore.'):
            pores = 'pore.' + pores
        pores = network.pores(pores)
    # Find coordinates of pores on given face
    coords = network['pore.coords'][pores]
    # Normalize the coordinates based on full network size
    c_norm = coords/network['pore.coords'].max(axis=0)
    # Identify axis of face by looking for dim with smallest delta
    diffs = np.amax(c_norm - np.average(c_norm, axis=0), axis=0)
    ax = np.where(diffs == diffs.min())[0][0]
    # Add new pore at center of domain
    new_coord = network['pore.coords'].mean(axis=0)
    domain_half_length = np.ptp(network['pore.coords'][:, ax])/2
    if coords[:, ax].mean() < network['pore.coords'][:, ax].mean():
        new_coord[ax] = new_coord[ax] - domain_half_length*(1 + offset)
    if coords[:, ax].mean() > network['pore.coords'][:, ax].mean():
        new_coord[ax] = new_coord[ax] + domain_half_length*(1 + offset)
    # Ps = np.arange(network.Np, network.Np + 1)
    extend(network=network, coords=[new_coord], labels=['reservoir'])
    conns = [[P, network.Np-1] for P in pores]
    # Ts = np.arange(network.Nt, network.Nt + len(conns))
    extend(network=network, conns=conns, labels=['reservoir'])
    # Compute the geometrical properties of the reservoir pore and throats
    # Confirm if network has any geometry props on it
    props = {'throat.length', 'pore.diameter', 'throat.volume'}
    if len(set(network.keys()).intersection(props)) > 0:
        raise Exception('Geometrical properties should be moved to a '
                        + 'geometry object first')
        # or just do this?:  geo = Imported(network=network)
    network.add_model(propname='pore.diameter',
                      model=mods.geometry.pore_size.largest_sphere)
    network.add_model(propname='throat.diameter_temp',
                      model=mods.geometry.throat_size.from_neighbor_pores,
                      mode='min')
    network.add_model(propname='throat.diameter',
                      model=mods.misc.scaled,
                      prop='throat.diameter_temp', factor=0.5)
    network.add_model(propname='throat.volume',
                      model=mods.geometry.throat_volume.cylinder)
    return network.project



import logging
import numpy as np
from openpnm.algorithms import ReactiveTransport
from openpnm.utils import Docorator


__all__ = ['AdvectionDiffusion']


docstr = Docorator()
logger = logging.getLogger(__name__)


@docstr.get_sections(base='AdvectionDiffusionSettings',
                     sections=['Parameters'])
@docstr.dedent
class AdvectionDiffusionSettings:
    r"""
    Parameters
    ----------
    %(ReactiveTransportSettings.parameters)s
    diffusive_conductance : str
        The name of the diffusive conductance values to be used by the
        specified ``conductance`` model to find the advective-diffusive
        conductance.
    hydraulic_conductance : str, optional
        The name of the hydraulic conductance values to be used by the
        specified ``conductance`` model to find the advective-diffusive
        conductance.
    pressure : str, optional
        The name of the pressure values calculated by the ``StokesFlow``
        algorithm.

    """
    quantity = 'pore.concentration'
    conductance = 'throat.ad_dif_conductance'
    diffusive_conductance = 'throat.diffusive_conductance'
    hydraulic_conductance = 'throat.hydraulic_conductance'
    pressure = 'pore.pressure'
    cache = False


class AdvectionDiffusion(ReactiveTransport):
    r"""
    A subclass of ReactiveTransport to simulate advection-diffusion
    """

    def __init__(self, name='ad_?', **kwargs):
        super().__init__(name=name, **kwargs)
        self.settings._update(AdvectionDiffusionSettings())
        self['pore.bc.outflow'] = np.nan

    def set_outflow_BC(self, pores, mode='add'):
        r"""
        Adds outflow boundary condition to the selected pores

        Parameters
        ----------
        pores : array_like
            The pore indices where the condition should be applied
        mode : str, optional
            Controls how the boundary conditions are applied. The default value
            is 'merge'. For definition of various modes, see the
            docstring for ``set_BC``.
        force : bool, optional
            If ``True`` then the ``'mode'`` is applied to all other bctypes as
            well. The default is ``False``.

        Notes
        -----
        Outflow condition means that the gradient of the solved quantity
        does not change, i.e. is 0.

        """
        pores = self._parse_indices(pores)

        # Calculating A[i,i] values to ensure the outflow condition
        network = self.project.network
        phase = self.project[self.settings['phase']]
        throats = network.find_neighbor_throats(pores=pores)
        C12 = network.conns[throats]
        P12 = phase[self.settings['pressure']][C12]
        gh = phase[self.settings['hydraulic_conductance']][throats]
        Q12 = -gh * np.diff(P12, axis=1).squeeze()
        Qp = np.zeros(self.Np)
        np.add.at(Qp, C12[:, 0], -Q12)
        np.add.at(Qp, C12[:, 1], Q12)

        self.set_BC(pores=pores, bcvalues=Qp[pores], bctype='outflow',
                    mode=mode)

    def _apply_BCs(self):
        r"""
        Applies Dirichlet, Neumann, and outflow BCs in order
        """
        # Apply Dirichlet and rate BCs
        super()._apply_BCs()
        if 'pore.bc.outflow' not in self.keys():
            return
        # Apply outflow BC
        diag = self.A.diagonal()
        ind = np.isfinite(self['pore.bc.outflow'])
        diag[ind] += self['pore.bc.outflow'][ind]
        self.A.setdiag(diag)


if __name__ == "__main__":
    import openpnm as op
    pn = op.network.Cubic(shape=[10, 10, 1])
    pn.add_model_collection(op.models.collections.geometry.spheres_and_cylinders)
    pn.regenerate_models()

    air = op.phase.Air(network=pn)
    air.add_model_collection(op.models.collections.physics.standard)
    air.regenerate_models()
    del air.models['throat.diffusive_conductance']
    del air.models['throat.hydraulic_conductance']
    air['throat.diffusive_conductance'] = 1e-15
    air['throat.hydraulic_conductance'] = 1e-15

    flow = op.algorithms.StokesFlow(network=pn, phase=air)
    flow.set_value_BC(pores=pn.pores('left'), values=1)
    flow.set_value_BC(pores=pn.pores('right'), values=0)
    flow.run()
    ad = op.algorithms.AdvectionDiffusion(network=pn, phase=air)
    ad.settings['cache'] = False
    ad.set_value_BC(pores=pn.pores('front'), values=1)
    ad.set_outflow_BC(pores=pn.pores('back'), mode='overwrite', force=False)
    ad.run()
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots(1, 1)
    ax.imshow(ad['pore.concentration'].reshape([10, 10]))



import numpy as np


__all__ = [
    'relax_points',
]


def relax_points(network, iterations=3, f=0.5):
    r"""
    Performs a relaxation of Delaunay points based on distance weighted average
    of connected Voronoi points

    Parameters
    ----------
    network : dict
        OpenPNM network dictionary
    iterations : int
        The number of time to iteratively apply the relaxation
    f : scalar
        A damping factor to limit the how much the point actually moves towards
        new computed point, to prevent overshooting.

    Returns
    -------
    coords : ndarray
        The Np-by-3 array of pore coordinates after relaxation.
    """
    if isinstance(f, int):
        f = [f]*iterations
    Ts = network.throats('interconnect')
    Ps = network.pores('delaunay')
    conns = network.conns[Ts]
    mask1 = network['pore.delaunay'][conns[:, 0]]
    mask2 = network['pore.delaunay'][conns[:, 1]]
    coords = np.copy(network.coords)
    for i in range(iterations):
        new_coords = np.zeros((len(Ps), 3), dtype=float)
        d = np.sqrt(np.sum((coords[conns[:, 0]] - coords[conns[:, 1]])**2, axis=1))
        weights = np.zeros(len(Ps), dtype=float)
        for ax in range(coords.shape[1]):
            np.add.at(new_coords[:, ax],
                      conns[:, 0][mask1],
                      d[mask1]*coords[conns[:, 1][mask1], ax])
            np.add.at(new_coords[:, ax],
                      conns[:, 1][mask2],
                      d[mask2]*coords[conns[:, 0][mask2], ax])
        np.add.at(weights, conns[:, 0][mask1], d[mask1])
        np.add.at(weights, conns[:, 1][mask2], d[mask2])
        delta = ((new_coords.T/weights).T - coords[Ps, :])*f[i]
        coords[Ps, :] += delta
    return coords



r"""
Physics
-------

This submodule contains models for calculating properties related to physical
transport processes, including conductances, reaction rates, and capillary
effects

"""

from . import ad_dif_conductance
from . import diffusive_conductance
from . import electrical_conductance
from . import hydraulic_conductance
from . import thermal_conductance
from . import source_terms
from . import capillary_pressure
from . import meniscus
from . import multiphase



import matplotlib.pyplot as plt
import numpy as np
from matplotlib import cm
from matplotlib import colors as mcolors
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.collections import LineCollection
from mpl_toolkits.mplot3d.art3d import Line3DCollection
from openpnm._skgraph.tools import dimensionality
from openpnm._skgraph.tools import get_node_prefix, get_edge_prefix


__all__ = [
    'plot_edges',
    'plot_nodes',
]


def plot_edges(network,
               edges=None,
               ax=None,
               size_by=None,
               color_by=None,
               cmap='jet',
               color='b',
               alpha=1.0,
               linestyle='solid',
               linewidth=1,
               **kwargs):  # pragma: no cover
    r"""
    Produce a 3D plot of the network topology

    This shows how edges connect for quick visualization without having
    to export data to veiw in Paraview.

    Parameters
    ----------
    network : dict
        The network dictionary
    edges : array_like (optional)
        The list of edges to plot if only a sub-sample is desired.  This is
        useful for inspecting a small region of the network.  If no edges are
        specified then all are shown.
    fig : Matplotlib figure handle and line property arguments (optional)
        If a ``fig`` is supplied, then the topology will be overlaid on this
        plot.  This makes it possible to combine coordinates and connections,
        and to color edges differently for instance.
    size_by : array_like (optional)
        An ndarray of edge values (e.g. alg['throat.rate']).  These
        values are used to scale the ``linewidth``, so if the lines are too
        thin, then increase ``linewidth``.
    color_by : str or array_like (optional)
        An ndarray of edge values (e.g. alg['edge.rate']).
    cmap : str or cmap object (optional)
        The matplotlib colormap to use if specfying an edge property
        for ``color_by``
    color : str, optional (optional)
        A matplotlib named color (e.g. 'r' for red).
    alpha : float (optional)
        The transparency of the lines, with 1 being solid and 0 being invisible
    linestyle : str (optional)
        Can be one of {'solid', 'dashed', 'dashdot', 'dotted'}.  Default is
        'solid'.
    linewidth : float (optional)
        Controls the thickness of drawn lines.  Is used to scale the thickness
        if ``size_by`` is given. Default is 1. If a value is provided for
        ``size_by`` then they are used to scale the ``linewidth``.
    **kwargs : dict
        All other keyword arguments are passed on to the ``Line3DCollection``
        class of matplotlib, so check their documentation for additional
        formatting options.

    Returns
    -------
    lc : LineCollection or Line3DCollection
        Matplotlib object containing the lines representing the throats.

    Notes
    -----
    To create a single plot containing both coordinates and connections,
    consider creating an empty figure and then passing the ``ax`` object as
    an argument to ``plot_connections`` and ``plot_coordinates``.
    Otherwise, each call to either of these methods creates a new figure.

    See Also
    --------
    plot_coordinates

    """
    node_prefix = get_node_prefix(network)
    edge_prefix = get_edge_prefix(network)
    conns = network[edge_prefix+'.conns']
    coords = network[node_prefix+'.coords']
    Ts = np.arange(conns.shape[0]) if edges is None else edges
    dim = dimensionality(network)
    ThreeD = True if dim.sum() == 3 else False
    # Add a dummy axis for 1D networks
    if dim.sum() == 1:
        dim[np.argwhere(~dim)[0]] = True

    if "fig" in kwargs.keys():
        raise Exception("'fig' argument is deprecated, use 'ax' instead.")
    if ax is None:
        fig, ax = plt.subplots()
    else:
        # The next line is necessary if ax was created using plt.subplots()
        fig, ax = ax.get_figure(), ax.get_figure().gca()
    if ThreeD and ax.name != '3d':
        fig.delaxes(ax)
        ax = fig.add_subplot(111, projection='3d')

    # Collect coordinates
    Ps = np.unique(conns[Ts])
    X, Y, Z = coords[Ps].T
    xyz = coords[:, dim]
    P1, P2 = conns[Ts].T
    throat_pos = np.column_stack((xyz[P1], xyz[P2])).reshape((Ts.size, 2, dim.sum()))

    # Deal with optional style related arguments
    if 'c' in kwargs.keys():
        color = kwargs.pop('c')
    color = mcolors.to_rgb(color) + tuple([alpha])
    # Override colors with color_by if given
    if color_by is not None:
        color = cm.get_cmap(name=cmap)(color_by / color_by.max())
        color[:, 3] = alpha
    if size_by is not None:
        linewidth = size_by / size_by.max() * linewidth

    if ThreeD:
        lc = Line3DCollection(throat_pos, colors=color, cmap=cmap,
                              linestyles=linestyle, linewidths=linewidth,
                              antialiaseds=np.ones_like(Ts), **kwargs)
    else:
        lc = LineCollection(throat_pos, colors=color, cmap=cmap,
                            linestyles=linestyle, linewidths=linewidth,
                            antialiaseds=np.ones_like(Ts), **kwargs)
    ax.add_collection(lc)

    _scale_axes(ax=ax, X=X, Y=Y, Z=Z)
    _label_axes(ax=ax, X=X, Y=Y, Z=Z)
    fig.tight_layout()

    return lc


def plot_nodes(network,
               nodes=None,
               ax=None,
               size_by=None,
               color_by=None,
               cmap='jet',
               color='r',
               alpha=1.0,
               marker='o',
               markersize=10,
               **kwargs):  # pragma: no cover
    r"""
    Produce a 3D plot showing specified nodecoordinates as markers.

    Parameters
    ----------
    network : dict
        The network dictionary
    nodes : array_like (optional)
        The list of nodes to plot if only a sub-sample is desired. This is
        useful for inspecting a small region of the network. If no nodes
        are specified then all are shown.
    ax : Matplotlib axis handle
        If ``ax`` is supplied, then the coordinates will be overlaid.
        This enables the plotting of multiple different sets of nodes as
        well as edge connections from ``plot_connections``.
    size_by : array_like
        An ndarray of node values (e.g. alg['node.radius']). These
        values are normalized by scaled by ``markersize``.
    color_by : array_like
        An ndarray of node values (e.g. alg['node.radius']).
    cmap : str or cmap object
        The matplotlib colormap to use if specfying a node property
        for ``color_by``
    color : str
        A matplotlib named color (e.g. 'r' for red).
    alpha : float
        The transparency of the lines, with 1 being solid and 0 being invisible
    marker : 's'
        The marker to use.  The default is a circle.  Options are explained
        `here <https://matplotlib.org/3.2.1/api/markers_api.html>`_
    markersize : scalar
        Controls size of marker, default is 1.0.  This value is used to scale
        the ``size_by`` argument if given.
    **kwargs
        All other keyword arguments are passed on to the ``scatter``
        function of matplotlib, so check their documentation for additional
        formatting options.

    Returns
    -------
    pc : PathCollection
        Matplotlib object containing the markers representing the pores.

    Notes
    -----
    To create a single plot containing both coordinates and connections,
    consider creating an empty figure and then passing the ``ax`` object as
    an argument to ``plot_edges`` and ``plot_nodes``.
    Otherwise, each call to either of these methods creates a new figure.

    See Also
    --------
    plot_edges

    """
    node_prefix = get_node_prefix(network)
    coords = network[node_prefix+'.coords']
    Ps = np.arange(coords.shape[0]) if nodes is None else nodes
    dim = dimensionality(network)
    ThreeD = True if dim.sum() == 3 else False
    # Add a dummy axis for 1D networks
    if dim.sum() == 1:
        dim[np.argwhere(~dim)[0]] = True
    # Add 2 dummy axes for 0D networks (1 pore only)
    if dim.sum() == 0:
        dim[[0, 1]] = True

    if "fig" in kwargs.keys():
        raise Exception("'fig' argument is deprecated, use 'ax' instead.")
    if ax is None:
        fig, ax = plt.subplots()
    else:
        # The next line is necessary if ax was created using plt.subplots()
        fig, ax = ax.get_figure(), ax.get_figure().gca()
    if ThreeD and ax.name != '3d':
        fig.delaxes(ax)
        ax = fig.add_subplot(111, projection='3d')

    # Collect specified coordinates
    X, Y, Z = coords[Ps].T
    # The bounding box for fig is the entire ntwork (to fix the problem with
    # overwriting figures' axes lim)
    Xl, Yl, Zl = coords.T

    # Parse formatting kwargs
    if 'c' in kwargs.keys():
        color = kwargs.pop('c')
    if 's' in kwargs.keys():
        markersize = kwargs.pop('s')
    if color_by is not None:
        color = cm.get_cmap(name=cmap)(color_by / color_by.max())
    if size_by is not None:
        markersize = size_by / size_by.max() * markersize

    if ThreeD:
        sc = ax.scatter(X, Y, Z, c=color, s=markersize, marker=marker,
                        alpha=alpha, **kwargs)
        _scale_axes(ax=ax, X=Xl, Y=Yl, Z=Zl)
    else:
        _X, _Y = np.column_stack((X, Y, Z))[:, dim].T
        sc = ax.scatter(_X, _Y, c=color, s=markersize, marker=marker,
                        alpha=alpha, **kwargs)
        _scale_axes(ax=ax, X=Xl, Y=Yl, Z=np.zeros_like(Yl))

    _label_axes(ax=ax, X=Xl, Y=Yl, Z=Zl)
    fig.tight_layout()

    return sc


def _label_axes(ax, X, Y, Z):
    labels = ["X", "Y", "Z"]
    dim = np.zeros(3, dtype=bool)
    for i, arr in enumerate([X, Y, Z]):
        if np.unique(arr).size > 1:
            dim[i] = True
    # Add a dummy axis for 1D networks
    if dim.sum() == 1:
        dim[np.argwhere(~dim)[0]] = True
    # Add 2 dummy axes for 0D networks (1 pore only)
    if dim.sum() == 0:
        dim[[0, 1]] = True
    dim_idx = np.argwhere(dim).squeeze()
    ax.set_xlabel(labels[dim_idx[0]])
    ax.set_ylabel(labels[dim_idx[1]])
    if hasattr(ax, "set_zlim"):
        ax.set_zlabel("Z")


def _scale_axes(ax, X, Y, Z):
    max_range = np.ptp([X, Y, Z]).max() / 2
    mid_x = (X.max() + X.min()) * 0.5
    mid_y = (Y.max() + Y.min()) * 0.5
    mid_z = (Z.max() + Z.min()) * 0.5
    ax.set_xlim(mid_x - max_range, mid_x + max_range)
    ax.set_ylim(mid_y - max_range, mid_y + max_range)
    if hasattr(ax, "set_zlim"):
        ax.set_zlim(mid_z - max_range, mid_z + max_range)
    else:
        ax.axis("equal")



import logging
from openpnm.utils import NestedDict
from openpnm.utils._misc import is_transient, nbr_to_str


logger = logging.getLogger(__name__)


def project_to_dict(project, categorize_by=['name'], flatten=False, element=None,
                    delim=' | '):
    r"""
    Returns a single dictionary object containing data from the given
    OpenPNM project, with the keys organized differently depending on
    optional arguments.

    Parameters
    ----------
    project : list
        An OpenPNM project object
    categorize_by : str or list[str]
        Indicates how the dictionaries should be organized.  The list can
        contain any, all or none of the following strings:

        **'object'** : If specified the dictionary keys will be stored
        under a general level corresponding to their type (e.g.
        'network/net_01/pore.all').

        **'name'** : If specified, then the data arrays are additionally
        categorized by their name.  This is enabled by default.

        **'data'** : If specified the data arrays are additionally
        categorized by ``label`` and ``property`` to separate *boolean*
        from *numeric* data.

        **'element'** : If specified the data arrays are
        additionally categorized by ``pore`` and ``throat``, meaning
        that the propnames are no longer prepended by a 'pore.' or
        'throat.'

    Returns
    -------
    A dictionary with the data stored in a hierarchical data structure, the
    actual format of which depends on the arguments to the function.

    """
    network = project.network
    phases = project.phases
    algs = project.algorithms

    if flatten:
        d = {}
    else:
        d = NestedDict(delimiter=delim)

    def build_path(obj, key):
        propname = key
        name = ''
        prefix = ''
        datatype = ''
        arr = obj[key]
        if 'object' in categorize_by:
            if hasattr(obj, 'coords'):
                prefix = 'network' + delim
            else:
                prefix = 'phase' + delim
        if 'element' in categorize_by:
            propname = key.replace('.', delim)
        if 'data' in categorize_by:
            if arr.dtype == bool:
                datatype = 'labels' + delim
            else:
                datatype = 'properties' + delim
        if 'name' in categorize_by:
            name = obj.name + delim
        path = prefix + name + datatype + propname
        return path

    for key in network.props(element=element) + network.labels(element=element):
        path = build_path(obj=network, key=key)
        d[path] = network[key]

    for phase in phases:
        for key in phase.props(element=element) + phase.labels(element=element):
            path = build_path(obj=phase, key=key)
            d[path] = phase[key]

    for alg in algs:
        try:  # 'quantity' is missing for multiphysics algorithm
            key = alg.settings['quantity']
        except AttributeError:
            break
        # only for transient algs
        transient = is_transient(alg)
        path = build_path(alg, key)
        if transient:
            times = alg.soln[key].t
            for i, t in enumerate(times):
                d[path + '#' + nbr_to_str(t)] = alg.soln[key][:, i]
        else:
            d[path] = alg.soln[key]

    return d



import numpy as _np
import openpnm.models.geometry.conduit_lengths as _conduit_lengths
from openpnm.models.geometry import _geodocs

__all__ = [
    "spheres_and_cylinders",
    "circles_and_rectangles",
    "cones_and_cylinders",
    "intersecting_cones",
    "hybrid_cones_and_cylinders",
    "trapezoids_and_rectangles",
    "hybrid_trapezoids_and_rectangles",
    "intersecting_trapezoids",
    "pyramids_and_cuboids",
    "intersecting_pyramids",
    "hybrid_pyramids_and_cuboids",
    "cubes_and_cuboids",
    "squares_and_rectangles",
    "intersecting_trapezoids",
    "ncylinders_in_series"
]


@_geodocs
def spheres_and_cylinders(
    network,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
):
    r"""
    Computes diffusive shape coefficient for conduits assuming pores are
    spheres and throats are cylinders.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------
    size_factors : ndarray
        Array (Nt by 3) containing conduit values for each element
        of the pore-throat-pore conduits. The array is formatted as
        ``[pore1, throat, pore2]``.

    Notes
    -----
    The diffusive size factor is the geometrical part of the pre-factor in
    Fick's law:

    .. math::

        n_A = \frac{A}{L} \Delta C_A
            = S_{diffusive} D_{AB} \Delta C_A

    Thus :math:`S_{diffusive}` represents the combined effect of the area and
    length of the *conduit*, which consists of a throat and 1/2 of the pore
    on each end.

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[1]).T
    L1, Lt, L2 = _conduit_lengths.spheres_and_cylinders(
        network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter
    ).T

    # Fi is the integral of (1/A) dx, x = [0, Li]
    F1 = 2 / (D1 * _np.pi) * _np.arctanh(2 * L1 / D1)
    F2 = 2 / (D2 * _np.pi) * _np.arctanh(2 * L2 / D2)
    Ft = Lt / (_np.pi / 4 * Dt ** 2)

    vals = _np.vstack([1/F1, 1/Ft, 1/F2]).T
    return vals


@_geodocs
def circles_and_rectangles(
    network,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
):
    r"""
    Computes diffusive shape coefficient for conduits assuming pores are
    circles and throats are rectangles

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[1]).T
    L1, Lt, L2 = _conduit_lengths.circles_and_rectangles(
        network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter
    ).T

    # Fi is the integral of (1/A) dx, x = [0, Li]
    F1 = 0.5 * _np.arcsin(2 * L1 / D1)
    F2 = 0.5 * _np.arcsin(2 * L2 / D2)
    Ft = Lt / Dt

    vals = _np.vstack([1/F1, 1/Ft, 1/F2]).T
    return vals


@_geodocs
def cones_and_cylinders(
    network,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
):
    r"""
    Computes diffusive shape coefficient assuming pores are truncated cones
    and throats are cylinders.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[1]).T
    L1, Lt, L2 = _conduit_lengths.cones_and_cylinders(
        network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter
    ).T

    # Fi is the integral of (1/A) dx, x = [0, Li]
    F1 = 4 * L1 / (D1 * Dt * _np.pi)
    F2 = 4 * L2 / (D2 * Dt * _np.pi)
    Ft = Lt / (_np.pi * Dt**2 / 4)

    vals = _np.vstack([1/F1, 1/Ft, 1/F2]).T
    return vals


@_geodocs
def intersecting_cones(
    network,
    pore_diameter="pore.diameter",
    throat_coords="throat.coords"
):
    r"""
    Computes diffusive shape coefficient assuming pores are intersecting cones.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Tcoords)s

    Returns
    -------

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[1]).T
    L1, Lt, L2 = _conduit_lengths.intersecting_cones(
        network,
        throat_coords=throat_coords
    ).T

    # Fi is the integral of (1/A) dx, x = [0, Li]
    F1 = 4 * L1 / (D1 * Dt * _np.pi)
    F2 = 4 * L2 / (D2 * Dt * _np.pi)
    inv_F_t = _np.full(len(Lt), _np.inf)

    vals = _np.vstack([1/F1, inv_F_t, 1/F2]).T
    return vals


@_geodocs
def hybrid_cones_and_cylinders(
    network,
    pore_diameter="pore.diameter",
    throat_coords="throat.coords"
):
    r"""
    Computes diffusive shape coefficient assuming pores are truncated cones
    and throats are cylinders.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Tcoords)s

    Returns
    -------

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[1]).T
    L1, Lt, L2 = _conduit_lengths.hybrid_cones_and_cylinders(
        network,
        pore_diameter=pore_diameter,
        throat_coords=throat_coords
    ).T

    # Fi is the integral of (1/A) dx, x = [0, Li]
    F1 = 4 * L1 / (D1 * Dt * _np.pi)
    F2 = 4 * L2 / (D2 * Dt * _np.pi)
    Ft = Lt / (_np.pi * Dt**2 / 4)
    mask = Lt == 0.0
    if mask.any():
        inv_F_t = _np.zeros(len(Ft))
        inv_F_t[~mask] = 1/Ft[~mask]
        inv_F_t[mask] = _np.inf
    else:
        inv_F_t = 1/Ft

    vals = _np.vstack([1/F1, inv_F_t, 1/F2]).T
    return vals


@_geodocs
def trapezoids_and_rectangles(
    network,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
):
    r"""
    Compute diffusive shape coefficient for conduits assuming pores are
    trapezoids and throats are rectangles.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[-1]).T
    L1, Lt, L2 = _conduit_lengths.trapezoids_and_rectangles(
        network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter
    ).T

    # Fi is the integral of (1/A) dx, x = [0, Li]
    F1 = L1 * _np.log(Dt / D1) / (Dt - D1)
    F2 = L2 * _np.log(Dt / D2) / (Dt - D2)
    Ft = Lt / Dt

    # Edge case where Di = Dt
    mask = _np.isclose(D1, Dt)
    F1[mask] = (L1 / D1)[mask]
    mask = _np.isclose(D2, Dt)
    F2[mask] = (L2 / D2)[mask]

    vals = _np.vstack([1/F1, 1/Ft, 1/F2]).T
    return vals


@_geodocs
def intersecting_trapezoids(
    network,
    pore_diameter="pore.diameter",
    throat_coords="throat.coords"
):
    r"""
    Computes diffusive shape coefficient for conduits of intersecting
    trapezoids.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Tcoords)s

    Returns
    -------

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[-1]).T
    L1, Lt, L2 = _conduit_lengths.intersecting_trapezoids(
        network,
        throat_coords=throat_coords
    ).T

    # Fi is the integral of (1/A) dx, x = [0, Li]
    F1 = L1 * _np.log(Dt / D1) / (Dt - D1)
    F2 = L2 * _np.log(Dt / D2) / (Dt - D2)

    # Edge case where Di = Dt
    mask = _np.isclose(D1, Dt)
    F1[mask] = (L1 / D1)[mask]
    mask = _np.isclose(D2, Dt)
    F2[mask] = (L2 / D2)[mask]
    inv_F_t = _np.full(len(Lt), _np.inf)

    vals = _np.vstack([1/F1, inv_F_t, 1/F2]).T
    return vals


@_geodocs
def hybrid_trapezoids_and_rectangles(
    network,
    pore_diameter="pore.diameter",
    throat_coords="throat.coords"
):
    r"""
    Compute diffusive shape coefficient for conduits assuming pores are
    trapezoids and throats are rectangles.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Tcoords)s

    Returns
    -------

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[-1]).T
    L1, Lt, L2 = _conduit_lengths.hybrid_trapezoids_and_rectangles(
        network,
        pore_diameter=pore_diameter,
        throat_coords=throat_coords
    ).T

    # Fi is the integral of (1/A) dx, x = [0, Li]
    F1 = L1 * _np.log(Dt / D1) / (Dt - D1)
    F2 = L2 * _np.log(Dt / D2) / (Dt - D2)
    Ft = Lt / Dt

    # Edge case where Di = Dt
    mask = _np.isclose(D1, Dt)
    F1[mask] = (L1 / D1)[mask]
    mask = _np.isclose(D2, Dt)
    F2[mask] = (L2 / D2)[mask]

    mask = Lt == 0.0
    if mask.any():
        inv_F_t = _np.zeros(len(Ft))
        inv_F_t[~mask] = 1/Ft[~mask]
        inv_F_t[mask] = _np.inf
    else:
        inv_F_t = 1/Ft

    vals = _np.vstack([1/F1, inv_F_t, 1/F2]).T
    return vals


@_geodocs
def pyramids_and_cuboids(
    network,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
):
    r"""
    Computes diffusive size factor for conduits of truncated pyramids and
    cuboids.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[-1]).T
    L1, Lt, L2 = _conduit_lengths.pyramids_and_cuboids(
        network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter
    ).T

    # Fi is the integral of (1/A) dx, x = [0, Li]
    F1 = L1 / (D1 * Dt)
    F2 = L2 / (D2 * Dt)
    Ft = Lt / Dt**2

    vals = _np.vstack([1/F1, 1/Ft, 1/F2]).T
    return vals


@_geodocs
def hybrid_pyramids_and_cuboids(
    network,
    pore_diameter="pore.diameter",
    throat_coords="throat.coords"
):
    r"""
    Computes diffusive size factor for conduits of truncated pyramids and
    cuboids.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Tcoords)s

    Returns
    -------

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[-1]).T
    L1, Lt, L2 = _conduit_lengths.hybrid_pyramids_and_cuboids(
        network,
        pore_diameter=pore_diameter,
        throat_coords=throat_coords
    ).T

    # Fi is the integral of (1/A) dx, x = [0, Li]
    F1 = L1 / (D1 * Dt)
    F2 = L2 / (D2 * Dt)
    Ft = Lt / Dt**2
    mask = Lt == 0.0
    if mask.any():
        inv_F_t = _np.zeros(len(Ft))
        inv_F_t[~mask] = 1/Ft[~mask]
        inv_F_t[mask] = _np.inf
    else:
        inv_F_t = 1/Ft

    vals = _np.vstack([1/F1, inv_F_t, 1/F2]).T
    return vals


@_geodocs
def intersecting_pyramids(
    network,
    pore_diameter="pore.diameter",
    throat_coords="throat.coords"
):
    r"""
    Computes diffusive size factor for conduits of pores with intersecting pyramids.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Tcoords)s

    Returns
    -------

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[-1]).T
    L1, Lt, L2 = _conduit_lengths.intersecting_pyramids(
        network,
        throat_coords=throat_coords
    ).T

    # Fi is the integral of (1/A) dx, x = [0, Li]
    F1 = L1 / (D1 * Dt)
    F2 = L2 / (D2 * Dt)
    inv_F_t = _np.full(len(Lt), _np.inf)

    vals = _np.vstack([1/F1, inv_F_t, 1/F2]).T
    return vals


@_geodocs
def cubes_and_cuboids(
    network,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
    pore_aspect=[1, 1, 1],
    throat_aspect=[1, 1, 1],
):
    r"""
    Computes diffusive shape coefficient for conduits assuming pores are
    cubes and throats are cuboids.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s
    pore_aspect : list
        Aspect ratio of the pores
    throat_aspect : list
        Aspect ratio of the throats

    Returns
    -------

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[-1]).T
    L1, Lt, L2 = _conduit_lengths.cubes_and_cuboids(
        network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter
    ).T

    # Fi is the integral of (1/A) dx, x = [0, Li]
    F1 = L1 / D1**2
    F2 = L2 / D2**2
    Ft = Lt / Dt**2

    vals = _np.vstack([1/F1, 1/Ft, 1/F2]).T
    return vals


@_geodocs
def squares_and_rectangles(
    network,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
    pore_aspect=[1, 1],
    throat_aspect=[1, 1],
):
    r"""
    Computes diffusive size factor for conduits assuming pores are squares
    and throats are rectangles.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s
    pore_aspect : list
        Aspect ratio of the pores
    throat_aspect : list
        Aspect ratio of the throats

    Returns
    -------

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[-1]).T
    L1, Lt, L2 = _conduit_lengths.squares_and_rectangles(
        network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter
    ).T

    # Fi is the integral of (1/A) dx, x = [0, Li]
    F1 = L1 / D1
    F2 = L2 / D2
    Ft = Lt / Dt

    vals = _np.vstack([1/F1, 1/Ft, 1/F2]).T
    return vals


@_geodocs
def ncylinders_in_series(
    network,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
    n=5
):
    r"""
    Computes diffusive size factors for conduits of spheres and cylinders
    with the spheres approximated as N cylinders in series.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s
    n : int
        Number of cylindrical divisions for each pore and throat

    Returns
    -------

    Notes
    -----

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[-1]).T
    # Ensure throats are never bigger than connected pores
    Dt = _np.minimum(Dt, 0.99 * _np.minimum(D1, D2))
    L1, Lt, L2 = _conduit_lengths.spheres_and_cylinders(
        network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter
    ).T

    dL1 = _np.linspace(0, L1, num=n)
    dL2 = _np.linspace(0, L2, num=n)
    r1 = D1 / 2 * _np.sin(_np.arccos(dL1 / (D1 / 2)))
    r2 = D2 / 2 * _np.sin(_np.arccos(dL2 / (D2 / 2)))

    gtemp = (_np.pi * r1**2 / (L1 / n)).T
    F1 = 1 / _np.sum(1 / gtemp, axis=1)
    gtemp = (_np.pi * r2 ** 2 / (L2 / n)).T
    F2 = 1 / _np.sum(1 / gtemp, axis=1)
    Ft = (_np.pi * (Dt / 2)**2 / (Lt)).T

    vals = _np.vstack([F1, Ft, F2]).T
    return vals



from openpnm.network import Network
from openpnm.utils import Docorator
from openpnm.topotools import label_faces
from openpnm._skgraph.generators import voronoi


docstr = Docorator()
__all__ = ['Voronoi']


@docstr.dedent
class Voronoi(Network):
    r"""
    Random network formed by Voronoi tessellation of arbitrary base points.

    Parameters
    ----------
    points : array_like or int
        Can either be an N-by-D array of point coordinates which will be
        used directly, or a scalar value indicating the number of points to
        generate.  If 3D points are supplied, and a 2D shape is specified, then
        the z-coordinate is ignored.See *Notes* for more details.
    shape : array_like
        The size and shape of the domain:

        ========== ============================================================
        shape      result
        ========== ============================================================
        [x, y, z]  A 3D cubic domain of dimension x, y and z
        [x, y, 0]  A 2D square domain of size x by y
        ========== ============================================================

    trim : bool, optional
        If ``True`` then all Voronoi vertices laying outside the domain
        will be removed.
    reflect : bool, optional
        If ``True`` then the base points will be reflected across
        all the faces of the domain prior to performing the tessellation. This
        feature is best combined with ``trim=True`` to make nice flat faces
        on all sides of the domain.
    f : float
        The fraction of points which should be reflected.  The default is 1 which
        reflects all the points in the domain, but this can lead to a lot of
        unnecessary points, so setting to 0.1 or 0.2 helps speed, but risks that
        the tessellation may not have smooth faces if not enough points are
        reflected.
    relaxation : int
        The number of time to iteratively relax the base points by moving them to
        the centroid of their respective Voronoi hulls. The default it 0.

    %(Network.parameters)s

    Notes
    -----
    These points are used by the Voronoi tessellation, meaning these points will
    each lie in the center of a Voronoi cell, so they will not be the pore centers.
    The number of pores in the returned network will greater that the number of
    points supplied or requested.

    It is also possible to generate circular ``[r, 0]``, cylindrical ``[r, z]``, and
    spherical domains ``[r]``, but this feature does not quite work as desired.
    It does not produce a truly clean outer surface since the tessellation are
    conducted in cartesian coordinates, so the circular and spherical surfaces
    have artifacts. Scipy recently added the ability to do tessellations on
    spherical surfaces, for geological applications, but this is not flexible
    enough, yet.

    """

    def __init__(
        self,
        shape,
        points,
        trim=True,
        reflect=True,
        f=1,
        relaxation=0,
        **kwargs
    ):
        super().__init__(**kwargs)
        net, vor = voronoi(
            points=points,
            shape=shape,
            trim=trim,
            reflect=reflect,
            f=f,
            relaxation=relaxation,
            node_prefix='pore',
            edge_prefix='throat',
        )
        self.update(net)
        self._post_init()
        self.vor = vor



import numpy as np
from openpnm.models.phase.mixtures import mixing_rule
from openpnm.models.phase import _phasedocs


__all__ = [
    'water_correlation',
    'gas_pure_st',
    'gas_pure_gesmr',
    'gas_mixture_hz',
    'liquid_pure_ls',
    'liquid_mixture_xweighted',
]


@_phasedocs
def water_correlation(
    phase,
    T='pore.temperature',
    salinity='pore.salinity'
):
    r"""
    Calculates viscosity of pure water or seawater at atmospheric pressure.

    This correlation uses Eq. (22) given by Sharqawy et. al [1]. Values at
    temperature higher than the normal boiling temperature are calculated
    at the saturation pressure.

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(salinity)s

    Returns
    -------
    mu : ndarray
        Array containing viscosity of water or seawater in [kg/m.s] or [Pa.s]

    Notes
    -----
     T must be in K, and S in g of salt per kg of phase, or ppt (parts per
        thousand)
    VALIDITY: 273 < T < 453 K; 0 < S < 150 g/kg;
    ACCURACY: 1.5 percent

    References
    ----------
    [1] Sharqawy M. H., Lienhard J. H., and Zubair, S. M., Desalination and
        Water Treatment, 2010.

    """
    T = phase[T]
    if salinity in phase.keys():
        S = phase[salinity]
    else:
        S = 0
    TC = T-273.15
    S = S/1000
    a1 = 1.5700386464E-01
    a2 = 6.4992620050E+01
    a3 = -9.1296496657E+01
    a4 = 4.2844324477E-05
    mu_w = a4 + 1/(a1*(TC+a2)**2+a3)
    a5 = 1.5409136040E+00
    a6 = 1.9981117208E-02
    a7 = -9.5203865864E-05
    a8 = 7.9739318223E+00
    a9 = -7.5614568881E-02
    a10 = 4.7237011074E-04
    A = a5 + a6*T + a7*T**2
    B = a8 + a9*T + a10*T**2
    mu_sw = mu_w*(1 + A*S + B*S**2)
    value = mu_sw
    return value


@_phasedocs
def air_correlation(
    phase,
    T='pore.temperature',
    n_V='pore.molar_density',
):
    r"""
    Calculates the viscosity of air at given conditions

    This model uses the correlation from [1].

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(n_V)s

    Returns
    -------

    References
    ----------
    [1] I forget

    """
    raise Exception("This function does not work yet")
    # Get props from object
    T = phase[T]
    rho = phase[n_V]
    # Declare given constants
    MW = 28.9586  # g/mol
    Tc = 132.6312  # K
    rho_c = 10447.7  # mol/m3
    sigma = 0.36  # nm
    e_k = 103.3  # K
    # Compute a few definitions
    delta = rho/rho_c
    T_star = T/e_k
    tau = np.atleast_2d(Tc/T)
    # Declare the summation variables
    ind = np.atleast_2d([0, 1, 2, 3, 4]).T
    N_i = np.atleast_2d([10.72, 1.122, 0.002019, -8.878, -0.02916]).T
    t_i = np.atleast_2d([0.2, 0.05, 2.4, 0.6, 3.6]).T
    d_i = np.atleast_2d([1, 4, 9, 1, 8]).T
    b_i = np.atleast_2d([0.431, -0.4623, 0.08406, 0.005341, -0.00331]).T
    gamma_i = np.atleast_2d([0, 0, 0, 1, 1]).T
    # Start crunching numbers
    omega = np.exp(np.sum(b_i*(np.atleast_2d(np.log(T_star))**(ind)), axis=0))
    mu_o = 0.9266958*(MW*T)**0.5/((sigma**2) * omega)
    temp = N_i * (tau**t_i) * (delta**d_i) * np.exp(-gamma_i*(delta**gamma_i))
    mu_t = np.sum(temp, axis=0)
    mu = mu_o + mu_t
    return mu


@_phasedocs
def gas_pure_st(
    phase,
    T='pore.temperature',
    Tc='param.critical_temperature',
    Pc='param.critical_pressure',
    MW='param.molecular_weight',
):
    r"""
    Calculates the viscosity of a pure gas using the correlation in [1]

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(Tc)s
    %(Pc)s
    %(MW)s

    Returns
    -------

    References
    ----------
    [1] Stiel and Thodos

    """
    #  stiel-thodos
    T = phase[T]
    Tc = phase[Tc]
    Pc = phase[Pc]/101325
    MW = phase[MW]

    mu = np.zeros_like(T)
    Tr = T/Tc
    zeta = Tc**(1/6)/((MW**0.5)*(Pc**(2/3)))
    mu_hi = (17.78e-5*(4.58*Tr - 1.67)**0.625)/zeta
    mu_lo = (34e-5*Tr**0.94)/zeta
    mask = Tr > 1.5
    mu[mask] = mu_hi[mask]
    mu[~mask] = mu_lo[~mask]
    return mu/1000


@_phasedocs
def gas_pure_gesmr(
    phase,
    T='pore.temperature',
    Tc='param.critical_temperature',
    Pc='param.critical_pressure',
    MW='param.molecular_weight',
):
    r"""
    Calculates the viscosity of a pure gas using the correlation  [1]

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(Tc)s
    %(Pc)s
    %(MW)s

    Returns
    -------

    References
    ----------
    [1] Gharagheizi et al
    """
    T = phase[T]
    MW = phase[MW]
    Pc = phase[Pc]
    Tc = phase[Tc]
    Tr = T/Tc
    mu = (1e-5)*Pc*Tr + (0.091 - 0.477/MW)*T + \
        MW*((1e-5)*Pc - 8.0*(MW**2)/(T**2))*(10.7639/Tc - 4.1929/T)
    mu = mu*1e-7
    mu = np.clip(mu, a_min=1e-7, a_max=np.inf)
    return mu


@_phasedocs
def gas_mixture_hz(
    phase,
    mus='pore.viscosity.*',
    MWs='param.molecular_weight.*',
):
    r"""
    Computes the viscosity of a gas mixture using the correlation in [1]

    Parameters
    ----------
    %(phase)s
    %(mus)s
    %(MWs)s

    Returns
    -------
    mu : ndarray
        An ndarray of Np length containing the viscosity of the mixture in
        each pore

    References
    ----------
    [1] Herning and Zipperer
    """
    MWs = phase.get_comp_vals(MWs)
    mus = phase.get_comp_vals(mus)
    xs = phase['pore.mole_fraction']
    num = 0.0
    denom = 0.0
    for k in xs.keys():
        num += xs[k]*mus[k]*MWs[k]**0.5
        denom += xs[k]*MWs[k]**0.5
    mu = num/denom
    return mu


@_phasedocs
def liquid_pure_ls(
    phase,
    T='pore.temperature',
    MW='param.molecular_weight',
    Tc='param.critical_temperature',
    Pc='param.critical_pressure',
    omega='param.acentric_factor',
):
    r"""
    Computes the viscosity of a pure liquid using the correlation in [1]

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(MW)s
    %(Tc)s
    %(Pc)s
    %(omega)s

    Returns
    -------
    mu : ndarray
        An ndarray of Np length containing the viscosity of the mixture in
        each pore

    References
    ----------
    [1] Letsou and Stiel

    """
    T = phase[T]
    MW = phase[MW]
    Tc = phase[Tc]
    Pc = phase[Pc]
    omega = phase[omega]
    Tr = T/Tc
    zeta = 2173.424 * (Tc**(1/6))/((MW**(0.5))*(Pc**(2/3)))
    zeta0 = (1.5174 - 2.135*Tr + 0.75*(Tr**2))*1e-5
    zeta1 = (4.2552 - 7.674*Tr + 3.4*(Tr**2))*1e-5
    mu = (zeta0 + omega*zeta1)/zeta
    return mu


def liquid_mixture_xweighted(
    phase,
    prop='pore.viscosity.*',
    mode='logarithmic',
    power=1,
):
    return mixing_rule(phase=phase, prop=prop, mode=mode, power=power)


liquid_mixture_xweighted.__doc__ = mixing_rule.__doc__



import codecs
import os
import os.path
import sys
from distutils.util import convert_path

from setuptools import find_packages, setup

sys.path.append(os.getcwd())
ver_path = convert_path('openpnm/__version__.py')


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            ver = line.split(delim)[1].split(".")
            if "dev0" in ver:
                ver.remove("dev0")
            return ".".join(ver)
    else:
        raise RuntimeError("Unable to find version string.")


# Read the contents of README file
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='openpnm',
    description='A framework for conducting pore network modeling simulations '
    + 'of multiphase transport in porous materials',
    long_description=long_description,
    long_description_content_type='text/markdown',
    version=get_version(ver_path),
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Physics'
    ],
    packages=find_packages("."),
    install_requires=[
        'chemicals',
        'docrep',
        'h5py',
        'jsonschema',
        'matplotlib',
        'networkx',
        'numba',
        'numpy',
        'pandas',
        'pyamg',
        'rich',
        'scikit-image',
        'scipy',
        'sympy',
        'thermo',
        'tqdm',
        'transforms3d',
    ],
    author='OpenPNM Team',
    author_email='jgostick@uwaterloo.ca',
    download_url='https://github.com/PMEAL/OpenPNM/',
    url='http://openpnm.org',
    project_urls={
        'Documentation': 'https://openpnm.org',
        'Source': 'https://github.com/PMEAL/OpenPNM',
        'Tracker': 'https://github.com/PMEAL/OpenPNM/issues',
    },
)
